# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
from glob import glob
from pathlib import Path

from .utils.misc import hardlink_or_copy


class APTCache:
    '''
    Manage cache of APT packages
    '''

    def __init__(self, osbase):
        self._cache_dir = os.path.join(osbase.global_config.aptcache_dir, osbase.name)

    def merge_from_dir(self, tmp_cache_dir):
        '''
        Merge in packages from a temporary cache
        '''

        from random import choice
        from string import digits, ascii_lowercase

        Path(self._cache_dir).mkdir(parents=True, exist_ok=True)
        for pkg_fname in glob(os.path.join(tmp_cache_dir, '*.deb')):
            pkg_basename = os.path.basename(pkg_fname)
            pkg_cachepath = os.path.join(self._cache_dir, pkg_basename)

            if not os.path.isfile(pkg_cachepath):
                pkg_tmp_name = (
                    pkg_cachepath + '.tmp-' + ''.join(choice(ascii_lowercase + digits) for _ in range(8))
                )
                shutil.copy2(pkg_fname, pkg_tmp_name)
                try:
                    os.rename(pkg_tmp_name, pkg_cachepath)
                except OSError:
                    # maybe some other debspawn instance tried to add the package just now,
                    # in that case we give up
                    os.remove(pkg_tmp_name)

    def create_instance_cache(self, tmp_cache_dir):
        '''
        Copy the cache to a temporary location for use in a new container instance.
        '''

        Path(self._cache_dir).mkdir(parents=True, exist_ok=True)
        Path(tmp_cache_dir).mkdir(parents=True, exist_ok=True)

        for pkg_fname in glob(os.path.join(self._cache_dir, '*.deb')):
            pkg_cachepath = os.path.join(tmp_cache_dir, os.path.basename(pkg_fname))

            if not os.path.isfile(pkg_cachepath):
                hardlink_or_copy(pkg_fname, pkg_cachepath)

    def clear(self):
        '''
        Remove all cache contents.
        '''

        Path(self._cache_dir).mkdir(parents=True, exist_ok=True)
        cache_size = len(glob(os.path.join(self._cache_dir, '*.deb')))

        old_cache_dir = self._cache_dir.rstrip(os.sep) + '.old'
        os.rename(self._cache_dir, old_cache_dir)
        Path(self._cache_dir).mkdir(parents=True, exist_ok=True)

        shutil.rmtree(old_cache_dir)

        return cache_size

    def delete(self):
        '''
        Remove cache completely - only useful when removing a base image completely.
        '''
        shutil.rmtree(self._cache_dir)



#!/usr/bin/env python3

import os
import sys

from debspawn import cli

thisfile = __file__
if not os.path.isabs(thisfile):
    thisfile = os.path.normpath(os.path.join(os.getcwd(), thisfile))

sys.exit(cli.run(thisfile, sys.argv[1:]))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2022 Matthias Klumpp <matthias@tenstral.net>
# Copyright (C) 2012-2013 Paul Tagliamonte <paultag@debian.org>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import sys
import shlex
import subprocess

from .log import TwoStreamLogger


class SubprocessError(Exception):
    def __init__(self, out, err, ret, cmd):
        self.out = out
        self.err = err
        self.ret = ret
        self.cmd = cmd
        super(SubprocessError, self).__init__('%s: %d\n%s' % (str(self.cmd), self.ret, str(self.err)))

    def __str__(self):
        return '%s: %d\n%s' % (str(self.cmd), self.ret, str(self.err))


# Input may be a byte string, a unicode string, or a file-like object
def run_command(command, input=None):
    if not isinstance(command, list):
        command = shlex.split(command)

    if not input:
        input = None
    elif isinstance(input, str):
        input = input.encode('utf-8')
    elif not isinstance(input, bytes):
        input = input.read()

    try:
        pipe = subprocess.Popen(
            command,
            shell=False,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except OSError:
        return (None, None, -1)

    (output, stderr) = pipe.communicate(input=input)
    (output, stderr) = (c.decode('utf-8', errors='ignore') for c in (output, stderr))
    return (output, stderr, pipe.returncode)


def safe_run(cmd, input=None, expected=0):
    if not isinstance(expected, tuple):
        expected = (expected,)

    out, err, ret = run_command(cmd, input=input)

    if ret not in expected:
        raise SubprocessError(out, err, ret, cmd)

    return out, err, ret


def run_forwarded(command):
    '''
    Run a command, forwarding all output to the current stdout as well as to
    our build-logger in case we have one set previously.
    '''
    if not isinstance(command, list):
        command = shlex.split(command)

    if isinstance(sys.stdout, TwoStreamLogger):
        proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        # ensure output is written to our file as well as stdout (as sys.stdout may be a redirect)
        while True:
            line = proc.stdout.readline()
            if proc.poll() is not None:
                break
            sys.stdout.write(str(line, 'utf-8', 'replace'))
        return proc
    else:
        return subprocess.run(command, check=False)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import typing as T
import platform
import subprocess

from .utils import (
    safe_run,
    temp_dir,
    print_info,
    print_warn,
    print_error,
    run_forwarded,
)
from .injectpkg import PackageInjector
from .utils.env import unicode_allowed, colored_output_allowed
from .utils.command import run_command

__systemd_version = None


def systemd_version():
    global __systemd_version
    if __systemd_version:
        return __systemd_version

    __systemd_version = -1
    try:
        out, _, _ = safe_run(['systemd-nspawn', '--version'])
        parts = out.split(' ', 2)
        if len(parts) >= 2:
            __systemd_version = int(parts[1])
    except Exception as e:
        print_warn('Unable to determine systemd version: {}'.format(e))

    return __systemd_version


def systemd_detect_virt():
    vm_name = 'unknown'
    try:
        out, _, _ = run_command(['systemd-detect-virt'])
        vm_name = out.strip()
    except Exception as e:
        print_warn('Unable to determine current virtualization: {}'.format(e))

    return vm_name


def systemd_version_atleast(expected_version: int):
    v = systemd_version()
    # we always assume we are running the highest version,
    # if we failed to determine the right systemd version
    if v < 0:
        return True
    if v >= expected_version:
        return True
    return False


def get_nspawn_personality(osbase):
    '''
    Return the syszemd-nspawn container personality for the given combination
    of host architecture and base OS.
    This allows running x86 builds on amd64 machines.
    '''
    import fnmatch

    if platform.machine() == 'x86_64' and fnmatch.filter([osbase.arch], 'i?86'):
        return 'x86'
    return None


def _execute_sdnspawn(
    osbase,
    parameters,
    machine_name,
    *,
    boot: bool = False,
    allow_permissions: list[str] = None,
    syscall_filter: list[str] = None,
    env_vars: dict[str, str] = None,
    private_users: bool = False,
    nowait: bool = False,
) -> T.Union[subprocess.CompletedProcess, subprocess.Popen]:
    '''
    Execute systemd-nspawn with the given parameters.
    Mess around with cgroups if necessary.
    '''
    import sys

    if not allow_permissions:
        allow_permissions = []
    if not syscall_filter:
        syscall_filter = []
    if not env_vars:
        env_vars = {}

    capabilities = []
    full_dev_access = False
    full_proc_access = False
    ro_kmods_access = False
    kvm_access = False
    all_privileges = False
    for perm in allow_permissions:
        perm = perm.lower()
        if perm.startswith('cap_') or perm == 'all':
            if perm == 'all':
                capabilities.append(perm)
                print_warn('Container retains all privileges.')
                all_privileges = True
            else:
                capabilities.append(perm.upper())
        elif perm == 'full-dev':
            full_dev_access = True
        elif perm == 'full-proc':
            full_proc_access = True
        elif perm == 'read-kmods':
            ro_kmods_access = True
        elif perm == 'kvm':
            kvm_access = True
        else:
            print_info('Unknown allowed permission: {}'.format(perm))

    if (
        capabilities or full_dev_access or full_proc_access or kvm_access
    ) and not osbase.global_config.allow_unsafe_perms:
        print_error(
            'Configuration does not permit usage of additional and potentially dangerous permissions. Exiting.'
        )
        sys.exit(9)

    cmd = ['systemd-nspawn']
    cmd.extend(['-M', machine_name])
    if boot:
        # if we boot the container, we also register it with machinectl, otherwise
        # we run an unregistered container with the command as PID2
        cmd.append('-b')
        cmd.append('--notify-ready=yes')
    else:
        cmd.append('--register=no')
        cmd.append('-a')
    if private_users:
        cmd.append('-U')  # User namespaces with --private-users=pick --private-users-chown, if possible
    if full_dev_access:
        cmd.extend(['--bind', '/dev'])
        if systemd_version_atleast(244):
            cmd.append('--console=pipe')
        cmd.extend(['--property=DeviceAllow=block-* rw', '--property=DeviceAllow=char-* rw'])
    if kvm_access and not full_dev_access:
        if os.path.exists('/dev/kvm'):
            cmd.extend(['--bind', '/dev/kvm'])
            cmd.extend(['--property=DeviceAllow=/dev/kvm rw'])
        else:
            print_warn(
                'Access to KVM requested, but /dev/kvm does not exist on the host. Is virtualization supported?'
            )
    if full_proc_access:
        cmd.extend(['--bind', '/proc'])
        if not all_privileges:
            print_warn('Container has access to host /proc')
    if ro_kmods_access:
        cmd.extend(['--bind-ro', '/lib/modules/'])
        cmd.extend(['--bind-ro', '/boot/'])
    if capabilities:
        cmd.extend(['--capability', ','.join(capabilities)])
    if syscall_filter:
        cmd.extend(['--system-call-filter', ' '.join(syscall_filter)])

    for v_name, v_value in env_vars.items():
        cmd.extend(['-E', '{}={}'.format(v_name, v_value)])

    cmd.extend(parameters)

    if nowait:
        return subprocess.Popen(cmd, shell=False, stdin=subprocess.DEVNULL)
    else:
        return run_forwarded(cmd)


def nspawn_run_persist(
    osbase,
    base_dir,
    machine_name,
    chdir,
    command: T.Union[list[str], str] = None,
    flags: T.Union[list[str], str] = None,
    *,
    tmp_apt_cache_dir: str = None,
    pkginjector: PackageInjector = None,
    allowed: list[str] = None,
    syscall_filter: list[str] = None,
    env_vars: dict[str, str] = None,
    private_users: bool = False,
    boot: bool = False,
    verbose: bool = False,
):
    if isinstance(command, str):
        command = command.split(' ')
    elif not command:
        command = []
    if isinstance(flags, str):
        flags = flags.split(' ')
    elif not flags:
        flags = []

    personality = get_nspawn_personality(osbase)

    def run_nspawn_with_aptcache(aptcache_tmp_dir):
        params = [
            '--chdir={}'.format(chdir),
            '--link-journal=no',
        ]
        if aptcache_tmp_dir:
            params.append('--bind={}:/var/cache/apt/archives/'.format(aptcache_tmp_dir))
        if pkginjector and pkginjector.instance_repo_dir:
            params.append('--bind={}:/srv/extra-packages/'.format(pkginjector.instance_repo_dir))

        if personality:
            params.append('--personality={}'.format(personality))
        params.extend(flags)
        params.extend(['-{}D'.format('' if verbose else 'q'), base_dir])

        # nspawn can not run a command in a booted container on its own
        if not boot:
            params.extend(command)
        sdns_nowait = boot and command

        # ensure the temporary apt cache is up-to-date
        if aptcache_tmp_dir:
            osbase.aptcache.create_instance_cache(aptcache_tmp_dir)

        # run command in container
        ns_proc = _execute_sdnspawn(
            osbase,
            params,
            machine_name,
            allow_permissions=allowed,
            syscall_filter=syscall_filter,
            env_vars=env_vars,
            private_users=private_users,
            boot=boot,
            nowait=sdns_nowait,
        )

        if not sdns_nowait:
            ret = ns_proc.returncode
        else:
            try:
                import time

                # the container is (hopefully) running now, but let's check for that
                time_ac_start = time.time()
                container_booted = False
                while (time.time() - time_ac_start) < 60:
                    scisr_out, _, _ = run_command(
                        [
                            'systemd-run',
                            '-GP',
                            '--wait',
                            '-qM',
                            machine_name,
                            'systemctl',
                            'is-system-running',
                        ]
                    )

                    # check if we are actually running, try again later if not
                    if scisr_out.strip() in ('running', 'degraded'):
                        print()
                        container_booted = True
                        break
                    time.sleep(0.5)

                if container_booted:
                    sdr_cmd = [
                        'systemd-run',
                        '-GP',
                        '--wait',
                        '-qM',
                        machine_name,
                        '--working-directory',
                        chdir,
                    ] + command
                    proc = run_forwarded(sdr_cmd)
                    ret = proc.returncode
                else:
                    ret = 7
                    print_error('Timed out while waiting for the container to boot.')
            finally:
                run_forwarded(['machinectl', 'poweroff', machine_name])
                try:
                    ns_proc.wait(30)
                except subprocess.TimeoutExpired:
                    ns_proc.terminate()

        # archive APT cache, so future runs of this command are faster (unless disabled in configuration)
        if aptcache_tmp_dir:
            osbase.aptcache.merge_from_dir(aptcache_tmp_dir)

        return ret

    if not osbase.cache_packages:
        # APT package caching was explicitly disabled by the user
        ret = run_nspawn_with_aptcache(None)
    elif tmp_apt_cache_dir:
        # we will be reusing an externally provided temporary APT cache directory
        ret = run_nspawn_with_aptcache(tmp_apt_cache_dir)
    else:
        # we will create our own temporary APT cache dir
        with temp_dir('aptcache-' + machine_name) as aptcache_tmp:
            ret = run_nspawn_with_aptcache(aptcache_tmp)

    return ret


def nspawn_run_ephemeral(
    osbase,
    base_dir,
    machine_name,
    chdir,
    command: T.Union[list[str], str] = None,
    flags: T.Union[list[str], str] = None,
    allowed: list[str] = None,
    syscall_filter: list[str] = None,
    env_vars: dict[str, str] = None,
    private_users: bool = False,
    boot: bool = False,
):
    if isinstance(command, str):
        command = command.split(' ')
    elif not command:
        command = []
    if isinstance(flags, str):
        flags = flags.split(' ')
    elif not flags:
        flags = []

    personality = get_nspawn_personality(osbase)

    params = ['--chdir={}'.format(chdir), '--link-journal=no']
    if personality:
        params.append('--personality={}'.format(personality))
    params.extend(flags)
    params.extend(['-qxD', base_dir])
    params.extend(command)

    return _execute_sdnspawn(
        osbase,
        params,
        machine_name,
        allow_permissions=allowed,
        syscall_filter=syscall_filter,
        env_vars=env_vars,
        private_users=private_users,
    ).returncode


def nspawn_make_helper_cmd(flags, build_uid: int):
    if isinstance(flags, str):
        flags = flags.split(' ')

    cmd = ['/usr/lib/debspawn/dsrun']
    if not colored_output_allowed():
        cmd.append('--no-color')
    if not unicode_allowed():
        cmd.append('--no-unicode')
    if build_uid > 0:
        cmd.append('--buid={}'.format(build_uid))

    cmd.extend(flags)
    return cmd


def nspawn_run_helper_ephemeral(
    osbase,
    base_dir,
    machine_name,
    helper_flags,
    chdir='/tmp',
    *,
    build_uid: int,
    nspawn_flags: T.Union[list[str], str] = None,
    allowed: list[str] = None,
    env_vars: dict[str, str] = None,
    private_users: bool = False,
):
    cmd = nspawn_make_helper_cmd(helper_flags, build_uid)
    return nspawn_run_ephemeral(
        base_dir,
        machine_name,
        chdir,
        cmd,
        flags=nspawn_flags,
        allowed=allowed,
        env_vars=env_vars,
        private_users=private_users,
    )


def nspawn_run_helper_persist(
    osbase,
    base_dir,
    machine_name,
    helper_flags,
    chdir='/tmp',
    *,
    build_uid: int,
    nspawn_flags: T.Union[list[str], str] = None,
    tmp_apt_cache_dir=None,
    pkginjector=None,
    allowed: list[str] = None,
    syscall_filter: list[str] = None,
    env_vars: dict[str, str] = None,
    private_users: bool = False,
):
    cmd = nspawn_make_helper_cmd(helper_flags, build_uid)
    return nspawn_run_persist(
        osbase,
        base_dir,
        machine_name,
        chdir,
        cmd,
        nspawn_flags,
        tmp_apt_cache_dir=tmp_apt_cache_dir,
        pkginjector=pkginjector,
        allowed=allowed,
        syscall_filter=syscall_filter,
        env_vars=env_vars,
        private_users=private_users,
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import sys

from .env import unicode_allowed
from .misc import safe_copy


def console_supports_color():
    '''
    Returns True if the running system's terminal supports color, and False
    otherwise.
    '''

    is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
    return 'ANSICON' in os.environ or is_a_tty


def print_textbox(title, tl, hline, tr, vline, bl, br):
    def write_utf8(s):
        sys.stdout.buffer.write(s.encode('utf-8'))

    tlen = len(title)
    write_utf8('\n{}'.format(tl))
    write_utf8(hline * (10 + tlen))
    write_utf8('{}\n'.format(tr))

    write_utf8('{}  {}'.format(vline, title))
    write_utf8(' ' * 8)
    write_utf8('{}\n'.format(vline))

    write_utf8(bl)
    write_utf8(hline * (10 + tlen))
    write_utf8('{}\n'.format(br))

    sys.stdout.flush()


def print_header(title):
    if unicode_allowed():
        print_textbox(title, '╔', '═', '╗', '║', '╚', '╝')
    else:
        print_textbox(title, '+', '═', '+', '|', '+', '+')


def print_section(title):
    if unicode_allowed():
        print_textbox(title, '┌', '─', '┐', '│', '└', '┘')
    else:
        print_textbox(title, '+', '-', '+', '|', '+', '+')


def print_info(*arg):
    '''
    Prints an information message and ensures that it shows up on
    stdout immediately.
    '''
    print(*arg)
    sys.stdout.flush()


def print_warn(*arg):
    '''
    Prints an information message and ensures that it shows up on
    stdout immediately.
    '''
    if console_supports_color():
        print('\033[93m/!\\\033[0m', *arg)
    else:
        print('/!\\', *arg)
    sys.stdout.flush()


def print_error(*arg):
    '''
    Prints an information message and ensures that it shows up on
    stdout immediately.
    '''
    if console_supports_color():
        print('\033[91mERROR:\033[0m', *arg, file=sys.stderr)
    else:
        print('ERROR:', *arg, file=sys.stderr)
    sys.stderr.flush()


def print_bullet(*arg, large: bool = False, indent: int = 0):
    '''
    Prints a bullet point to the console, with a set
    indentation and style.
    '''
    if unicode_allowed():
        b = '●' if large else '•'
    else:
        b = '*'
    print((' ' * indent) + b, *arg)


def print_bool_item(prefix: str, b: bool, text_true: str = 'yes', text_false: str = 'no'):
    '''
    Prints a (colored, if possible) boolean item with a given prefix.
    '''
    if console_supports_color():
        s = '\033[92m{}\033[0m'.format(text_true) if b else '\033[91m{}\033[0m'.format(text_false)
    else:
        s = text_true if b else text_false
    if prefix:
        print(prefix, s)
    else:
        print(prefix)
    sys.stdout.flush()


def input_bool(question_text, default=False) -> bool:
    """As user a Yes/No question."""
    if default:
        default_info = '[Y/n]'
    else:
        default_info = '[y/N]'
    while True:
        try:
            in_str = input('{} {}:'.format(question_text, default_info))
        except EOFError:
            return default
        if in_str == 'y' or in_str == 'Y':
            return True
        elif in_str == 'n' or in_str == 'N':
            return False
        elif not in_str:
            return default


class TwoStreamLogger:
    '''
    Permits logging messages to stdout/stderr as well as to a file.
    '''

    class Buffer:
        def __init__(self, fstream, cstream):
            self._fstream = fstream
            self._cstream = cstream

        def write(self, message):
            self._fstream.write(str(message, 'utf-8', 'replace'))
            self._cstream.buffer.write(message)

    def __init__(self, fstream, cstream, fflush_always=False):
        self._fstream = fstream
        self._cstream = cstream
        self._fflush_always = fflush_always
        self._colorsub = re.compile('\x1b\\[(K|.*?m)')
        self.buffer = TwoStreamLogger.Buffer(fstream, cstream)

    def write(self, message):
        # write message to console
        self._cstream.write(message)
        if self._fflush_always:
            self.flush()

        # write message to file, stripping ANSI colors
        self._fstream.write(self._colorsub.sub('', message))

    def flush(self):
        self._cstream.flush()
        self._fstream.flush()

    def copy_to(self, fname):
        self.flush()
        safe_copy(self._fstream.name, fname, preserve_mtime=False)

    def isatty(self):
        return self._cstream.isatty()


def capture_console_output():
    '''
    Direct console output to a file as well
    as to the original stdout/stderr terminal.
    '''
    from tempfile import NamedTemporaryFile

    logfile = NamedTemporaryFile(mode='a', prefix='ds_', suffix='.log')
    nstdout = TwoStreamLogger(logfile, sys.stdout)
    nstderr = TwoStreamLogger(logfile, sys.stderr, True)

    sys.stdout = nstdout
    sys.stderr = nstderr


def save_captured_console_output(fname):
    from .env import get_owner_uid_gid

    if hasattr(sys.stdout, 'copy_to'):
        o_uid, o_gid = get_owner_uid_gid()
        sys.stdout.copy_to(fname)
        os.chown(fname, o_uid, o_gid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import platform
import subprocess
from glob import glob
from collections.abc import Iterable

from .nspawn import nspawn_run_persist, nspawn_run_helper_persist
from .injectpkg import PackageInjector
from .utils.env import (
    ensure_root,
    get_tree_size,
    get_free_space,
    get_owner_uid_gid,
    switch_unprivileged,
    get_random_free_uid_gid,
)
from .utils.log import (
    input_bool,
    print_info,
    print_warn,
    print_error,
    print_bullet,
    print_header,
    print_section,
    capture_console_output,
    save_captured_console_output,
)
from .utils.misc import (
    cd,
    listify,
    temp_dir,
    safe_copy,
    format_filesize,
    version_noepoch,
)
from .utils.command import safe_run


class BuildError(Exception):
    """Package build failed with a generic error."""


def interact_with_build_environment(
    osbase,
    instance_dir,
    machine_name,
    *,
    pkg_dir_root,
    source_pkg_dir,
    aptcache_tmp,
    pkginjector,
    prev_exitcode,
) -> bool:
    '''Launch an interactive shell in the build environment'''

    # find the right directory to switch to
    pkg_dir = pkg_dir_root
    for f in glob(os.path.join(pkg_dir, '*')):
        if os.path.isdir(f):
            pkg_dir = f
            break

    print()
    print_info('Launching interactive shell in build environment.')
    if prev_exitcode != 0:
        print_info('The previous build step failed with exit code {}'.format(prev_exitcode))
    else:
        print_info('The previous build step was successful.')
    print_info('Temporary location of package files on the host:\n  => file://{}'.format(pkg_dir))
    print_info('Press CTL+D to exit the interactive shell.')
    print()

    nspawn_flags = ['--bind={}:/srv/build/'.format(pkg_dir_root)]
    nspawn_run_persist(
        osbase,
        instance_dir,
        machine_name,
        chdir=os.path.join('/srv/build', os.path.basename(pkg_dir)),
        flags=nspawn_flags,
        tmp_apt_cache_dir=aptcache_tmp,
        pkginjector=pkginjector,
        syscall_filter=osbase.global_config.syscall_filter,
        verbose=True,
    )

    print()
    copy_artifacts = input_bool(
        'Should any generated build artifacts (binary/source packages, etc.) be saved?', default=False
    )
    if copy_artifacts:
        print_bullet('Artifacts will be copied to the results directory.')
    else:
        print_bullet('Artifacts will not be kept.')

    if source_pkg_dir:
        copy_changes = input_bool(
            (
                'Should changes to the debian/ directory be copied back to the host?\n'
                'This will OVERRIDE all changes made on files on the host.'
            ),
            default=False,
        )

        if copy_changes:
            print_info('Cleaning up...')
            # clean the source tree. we intentionally ignore errors here.
            nspawn_run_persist(
                osbase,
                instance_dir,
                machine_name,
                chdir=os.path.join('/srv/build', os.path.basename(pkg_dir)),
                flags=nspawn_flags,
                command=['dpkg-buildpackage', '-T', 'clean'],
                tmp_apt_cache_dir=aptcache_tmp,
                pkginjector=pkginjector,
            )

            print()
            print_info('Copying back changes...')
            known_files = {}
            dest_debian_dir = os.path.join(source_pkg_dir, 'debian')
            src_debian_dir = os.path.join(pkg_dir, 'debian')

            # get uid/gid of the user who invoked us
            o_uid, o_gid = get_owner_uid_gid()

            # collect list of existing packages
            for sdir, _, files in os.walk(dest_debian_dir):
                for f in files:
                    fname = os.path.join(sdir, f)
                    known_files[os.path.relpath(fname, dest_debian_dir)] = fname

            # walk through the source files, copying everything to the destination
            for sdir, _, files in os.walk(src_debian_dir):
                for f in files:
                    fname = os.path.join(sdir, f)
                    rel_fname = os.path.relpath(fname, src_debian_dir)
                    dest_fname = os.path.normpath(os.path.join(dest_debian_dir, rel_fname))
                    dest_dir = os.path.dirname(dest_fname)
                    if rel_fname in known_files:
                        del known_files[rel_fname]

                    if os.path.isdir(fname):
                        print('New dir: {}'.format(rel_fname))
                        with switch_unprivileged():
                            os.makedirs(dest_fname, exist_ok=True)
                        continue
                    if not os.path.isdir(dest_dir):
                        print('New dir: {}'.format(os.path.relpath(dest_dir, dest_debian_dir)))
                        with switch_unprivileged():
                            os.makedirs(dest_dir, exist_ok=True)

                    print('Copy: {}'.format(rel_fname))
                    safe_copy(fname, dest_fname)
                    os.chown(dest_fname, o_uid, o_gid, follow_symlinks=False)

            for rel_fname, fname in known_files.items():
                print('Delete: {}'.format(rel_fname))
                os.remove(fname)
            print()
        else:
            print_bullet('Discarding build environment.')
    else:
        print_info('Can not copy back changes as original package directory is unknown.')

    return copy_artifacts


def internal_execute_build(
    osbase,
    pkg_dir,
    build_only=None,
    *,
    qa_lintian=False,
    interact=False,
    source_pkg_dir=None,
    buildflags: list[str] = None,
    build_env: dict[str, str] = None,
):
    '''Perform the actual build on an extracted package directory'''
    assert not build_only or isinstance(build_only, str)
    if not pkg_dir:
        raise ValueError('Package directory is missing!')
    pkg_dir = os.path.normpath(pkg_dir)
    if not build_env:
        build_env = {}

    # get a fresh UID to give to our build user within the container
    builder_uid = get_random_free_uid_gid()[0]

    with osbase.new_instance() as (instance_dir, machine_name):
        # first, check basic requirements

        # instance dir and pkg dir are both temporary directories, so they
        # will be on the same filesystem configured as workspace for debspawn.
        # therefore we only check on directory.
        free_space = get_free_space(instance_dir)
        print_info('Free space in workspace: {}'.format(format_filesize(free_space)))

        # check for at least 512MiB - this is a ridiculously small amount, so the build will likely fail.
        # but with even less, even attempting a build is pointless.
        if (free_space / 2048) < 512:
            print_error('Not enough free space available in workspace.')
            return 8

        # prepare the build. At this point, we only run trusted code and the container
        # has network access
        with temp_dir('pkgsync-' + machine_name) as pkgsync_tmp:
            # create temporary locations set up and APT cache sharing and package injection
            aptcache_tmp = os.path.join(pkgsync_tmp, 'aptcache')
            pkginjector = PackageInjector(osbase)
            if pkginjector.has_injectables():
                pkginjector.create_instance_repo(os.path.join(pkgsync_tmp, 'pkginject'))

            # set up the build environment
            nspawn_flags = ['--bind={}:/srv/build/'.format(pkg_dir)]
            prep_flags = ['--build-prepare']

            # if we force a suite and have injected packages, the injected packages
            # will never be picked up.
            if not pkginjector.has_injectables():
                prep_flags.extend(['--suite', osbase.suite])

            if build_only == 'arch':
                prep_flags.append('--arch-only')
            r = nspawn_run_helper_persist(
                osbase,
                instance_dir,
                machine_name,
                prep_flags,
                '/srv',
                build_uid=builder_uid,
                nspawn_flags=nspawn_flags,
                tmp_apt_cache_dir=aptcache_tmp,
                pkginjector=pkginjector,
            )
            if r != 0:
                print_error('Build environment setup failed.')
                return False

            # run the actual build. At this point, code is less trusted, and we disable network access.
            nspawn_flags = ['--bind={}:/srv/build/'.format(pkg_dir), '-u', 'builder', '--private-network']
            helper_flags = ['--build-run']
            helper_flags.extend(['--suite', osbase.suite])
            if buildflags:
                helper_flags.append('--buildflags={}'.format(';'.join(buildflags)))
            r = nspawn_run_helper_persist(
                osbase,
                instance_dir,
                machine_name,
                helper_flags,
                '/srv',
                build_uid=builder_uid,
                nspawn_flags=nspawn_flags,
                tmp_apt_cache_dir=aptcache_tmp,
                pkginjector=pkginjector,
                env_vars=build_env,
                syscall_filter=osbase.global_config.syscall_filter,
            )
            # exit, unless we are in interactive mode
            if r != 0 and not interact:
                return False

            if qa_lintian and r == 0:
                # running Lintian was requested, so do so.
                # we use Lintian from the container, so we validate with the validator from
                # the OS the package was actually built against
                nspawn_flags = ['--bind={}:/srv/build/'.format(pkg_dir)]
                r = nspawn_run_helper_persist(
                    osbase,
                    instance_dir,
                    machine_name,
                    ['--run-qa', '--lintian'],
                    '/srv',
                    build_uid=builder_uid,
                    nspawn_flags=nspawn_flags,
                    tmp_apt_cache_dir=aptcache_tmp,
                    pkginjector=pkginjector,
                )
                if r != 0:
                    print_error('QA failed.')
                    return False
                print()  # extra blank line after Lintian output

            if interact:
                ri = interact_with_build_environment(
                    osbase,
                    instance_dir,
                    machine_name,
                    pkg_dir_root=pkg_dir,
                    source_pkg_dir=source_pkg_dir,
                    aptcache_tmp=aptcache_tmp,
                    pkginjector=pkginjector,
                    prev_exitcode=r,
                )
                # if we exit with a non-True result, we stop here and don't proceed
                # with the next steps that save artifacts.
                if not ri:
                    return False

            build_dir_size = get_tree_size(pkg_dir)
            print_info(
                'This build required {} of dedicated disk space.'.format(format_filesize(build_dir_size))
            )

    return True


def print_build_detail(osbase, pkgname, version):
    print_info('Package: {}'.format(pkgname))
    print_info('Version: {}'.format(version))
    print_info('Distribution: {}'.format(osbase.suite))
    print_info('Architecture: {}'.format(osbase.arch))
    print_info()


def _read_source_package_details():
    out, err, ret = safe_run(['dpkg-parsechangelog'])
    if ret != 0:
        raise BuildError('Running dpkg-parsechangelog failed: {}{}'.format(out, err))

    pkg_sourcename = None
    pkg_version = None
    for line in out.split('\n'):
        if line.startswith('Source: '):
            pkg_sourcename = line[8:].strip()
        elif line.startswith('Version: '):
            pkg_version = line[9:].strip()

    if not pkg_sourcename or not pkg_version:
        print_error('Unable to determine source package name or source package version. Can not continue.')
        return None, None, None

    pkg_version_dsc = version_noepoch(pkg_version)
    dsc_fname = '{}_{}.dsc'.format(pkg_sourcename, pkg_version_dsc)

    return pkg_sourcename, pkg_version, dsc_fname


def _get_build_flags(build_only=None, include_orig=False, maintainer=None, extra_flags: Iterable[str] = None):
    import shlex

    buildflags = []
    extra_flags = listify(extra_flags)

    if build_only:
        if build_only == 'binary':
            buildflags.append('-b')
        elif build_only == 'arch':
            buildflags.append('-B')
        elif build_only == 'indep':
            buildflags.append('-A')
        elif build_only == 'source':
            buildflags.append('-S')
        else:
            print_error('Invalid build-only flag "{}". Can not continue.'.format(build_only))
            return False, []

    if include_orig:
        buildflags.append('-sa')
    if maintainer:
        buildflags.append('-m{}'.format(maintainer.replace(';', ',')))
        buildflags.append('-e{}'.format(maintainer.replace(';', ',')))
    for flag_raw in extra_flags:
        buildflags.extend(shlex.split(flag_raw))

    return True, buildflags


def _sign_result(results_dir, spkg_name, spkg_version, build_arch):
    print_section('Signing Package')
    spkg_version_noepoch = version_noepoch(spkg_version)
    changes_basename = '{}_{}_{}.changes'.format(spkg_name, spkg_version_noepoch, build_arch)

    with switch_unprivileged():
        proc = subprocess.run(['debsign', os.path.join(results_dir, changes_basename)], check=False)
        if proc.returncode != 0:
            print_error('Signing failed.')
            return False
    return True


def _print_system_info():
    from . import __version__
    from .utils.misc import current_time_string

    print_info(
        'debspawn {version} on {host} at {time}'.format(
            version=__version__, host=platform.node(), time=current_time_string()
        )
    )


def build_from_directory(
    osbase,
    pkg_dir,
    *,
    sign=False,
    build_only=None,
    include_orig=False,
    maintainer=None,
    clean_source=False,
    qa_lintian=False,
    interact=False,
    log_build=True,
    extra_dpkg_flags: list[str] = None,
    build_env: dict[str, str] = None,
):
    ensure_root()
    osbase.ensure_exists()
    extra_dpkg_flags = listify(extra_dpkg_flags)

    if interact and log_build:
        print_warn('Build log and interactive mode can not be enabled at the same time. Disabling build log.')
        print()
        log_build = False

    # capture console output if we should log the build
    if log_build:
        capture_console_output()

    if not pkg_dir:
        pkg_dir = os.getcwd()
    pkg_dir = os.path.abspath(pkg_dir)

    r, buildflags = _get_build_flags(build_only, include_orig, maintainer, extra_dpkg_flags)
    if not r:
        return False

    _print_system_info()
    print_header('Package build (from directory)')

    print_section('Creating source package')
    with cd(pkg_dir):
        with switch_unprivileged():
            deb_files_fname = os.path.join(pkg_dir, 'debian', 'files')
            if os.path.isfile(deb_files_fname):
                deb_files_fname = None  # the file already existed, we don't need to clean it up later

            pkg_sourcename, pkg_version, dsc_fname = _read_source_package_details()
            if not pkg_sourcename:
                return False

            cmd = ['dpkg-buildpackage', '-S', '--no-sign']
            # d/rules clean requires build dependencies installed if run on the host
            # we avoid that by default, unless explicitly requested
            if not clean_source:
                cmd.append('-nc')

            proc = subprocess.run(cmd, check=False)
            if proc.returncode != 0:
                return False

            # remove d/files file that was created when generating the source package.
            # we only clean up the file if it didn't exist prior to us running the command.
            if deb_files_fname:
                try:
                    os.remove(deb_files_fname)
                except OSError:
                    pass

    print_header('Package build')
    print_build_detail(osbase, pkg_sourcename, pkg_version)

    success = False
    with temp_dir(pkg_sourcename) as pkg_tmp_dir:
        with cd(pkg_tmp_dir):
            cmd = ['dpkg-source', '-x', os.path.join(pkg_dir, '..', dsc_fname)]
            proc = subprocess.run(cmd, check=False)
            if proc.returncode != 0:
                return False

        success = internal_execute_build(
            osbase,
            pkg_tmp_dir,
            build_only,
            qa_lintian=qa_lintian,
            interact=interact,
            source_pkg_dir=pkg_dir,
            buildflags=buildflags,
            build_env=build_env,
        )

        # copy build results
        if success:
            osbase.retrieve_artifacts(pkg_tmp_dir)

    # save buildlog, if we generated one
    log_fname = os.path.join(
        osbase.results_dir,
        '{}_{}_{}.buildlog'.format(pkg_sourcename, version_noepoch(pkg_version), osbase.arch),
    )
    save_captured_console_output(log_fname)

    # exit, there is nothing more to do if no package was built
    if not success:
        return False

    # sign the resulting package
    if sign:
        r = _sign_result(osbase.results_dir, pkg_sourcename, pkg_version, osbase.arch)
        if not r:
            return False

    print_info('Done.')

    return True


def build_from_dsc(
    osbase,
    dsc_fname,
    *,
    sign=False,
    build_only=None,
    include_orig=False,
    maintainer=None,
    qa_lintian: bool = False,
    interact: bool = False,
    log_build: bool = True,
    extra_dpkg_flags: Iterable[str] = None,
    build_env: dict[str, str] = None,
):
    ensure_root()
    osbase.ensure_exists()
    extra_dpkg_flags = listify(extra_dpkg_flags)

    if interact and log_build:
        print_warn('Build log and interactive mode can not be enabled at the same time. Disabling build log.')
        print()
        log_build = False

    # capture console output if we should log the build
    if log_build:
        capture_console_output()

    r, buildflags = _get_build_flags(build_only, include_orig, maintainer, extra_dpkg_flags)
    if not r:
        return False

    _print_system_info()

    success = False
    dsc_fname = os.path.abspath(os.path.normpath(dsc_fname))
    tmp_prefix = os.path.basename(dsc_fname).replace('.dsc', '').replace(' ', '-')
    with temp_dir(tmp_prefix) as pkg_tmp_dir:
        with cd(pkg_tmp_dir):
            cmd = ['dpkg-source', '-x', dsc_fname]
            proc = subprocess.run(cmd, check=False)
            if proc.returncode != 0:
                return False

            pkg_srcdir = None
            for f in glob('./*'):
                if os.path.isdir(f):
                    pkg_srcdir = f
                    break
            if not pkg_srcdir:
                print_error('Unable to find source directory of extracted package.')
                return False

            with cd(pkg_srcdir):
                pkg_sourcename, pkg_version, dsc_fname = _read_source_package_details()
                if not pkg_sourcename:
                    return False

            print_header('Package build')
            print_build_detail(osbase, pkg_sourcename, pkg_version)

        success = internal_execute_build(
            osbase,
            pkg_tmp_dir,
            build_only,
            qa_lintian=qa_lintian,
            interact=interact,
            buildflags=buildflags,
            build_env=build_env,
        )

        # copy build results
        if success:
            osbase.retrieve_artifacts(pkg_tmp_dir)

    # save buildlog, if we generated one
    log_fname = os.path.join(
        osbase.results_dir,
        '{}_{}_{}.buildlog'.format(pkg_sourcename, version_noepoch(pkg_version), osbase.arch),
    )
    save_captured_console_output(log_fname)

    # build log is saved, but no artifacts are available, so there's nothing more to do
    if not success:
        return False

    # sign the resulting package
    if sign:
        r = _sign_result(osbase.results_dir, pkg_sourcename, pkg_version, osbase.arch)
        if not r:
            return False

    print_info('Done.')
    return True



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
from glob import glob
from pathlib import Path
from contextlib import contextmanager

from .utils import temp_dir, print_info, hardlink_or_copy


class PackageInjector:
    '''
    Inject packages from external sources into the container APT environment.
    '''

    def __init__(self, osbase):
        self._pkgs_basedir = osbase.global_config.injected_pkgs_dir
        self._pkgs_specific_dir = os.path.join(self._pkgs_basedir, osbase.name)
        self._has_injectables = None
        self._instance_repo_dir = None

    def has_injectables(self):
        '''Return True if we actually have any packages ready to inject'''

        if type(self._has_injectables) is bool:
            return self._has_injectables

        if os.path.exists(self._pkgs_basedir) and os.path.isdir(self._pkgs_basedir):
            for f in os.listdir(self._pkgs_basedir):
                if os.path.isfile(os.path.join(self._pkgs_basedir, f)):
                    self._has_injectables = True
                    return True

            if os.path.exists(self._pkgs_specific_dir) and os.path.isdir(self._pkgs_specific_dir):
                for f in os.listdir(self._pkgs_specific_dir):
                    if os.path.isfile(os.path.join(self._pkgs_specific_dir, f)):
                        self._has_injectables = True
                        return True

        self._has_injectables = False
        return False

    def create_instance_repo(self, tmp_repo_dir):
        '''
        Create a temporary location where all injected packages for this container
        are copied to.
        '''

        Path(self._pkgs_basedir).mkdir(parents=True, exist_ok=True)
        Path(tmp_repo_dir).mkdir(parents=True, exist_ok=True)

        print_info('Copying injected packages to instance location')
        self._instance_repo_dir = tmp_repo_dir

        # copy/link injected packages specific to this environment
        if os.path.isdir(self._pkgs_specific_dir):
            for pkg_fname in glob(os.path.join(self._pkgs_specific_dir, '*.deb')):
                pkg_path = os.path.join(tmp_repo_dir, os.path.basename(pkg_fname))

                if not os.path.isfile(pkg_path):
                    hardlink_or_copy(pkg_fname, pkg_path)

        # copy/link injected packages used by all environments
        for pkg_fname in glob(os.path.join(self._pkgs_basedir, '*.deb')):
            pkg_path = os.path.join(tmp_repo_dir, os.path.basename(pkg_fname))

            if not os.path.isfile(pkg_path):
                hardlink_or_copy(pkg_fname, pkg_path)

    @property
    def instance_repo_dir(self) -> str:
        return self._instance_repo_dir


@contextmanager
def package_injector(osbase, machine_name=None):
    '''
    Create a package injector as context manager and make
    it create a new temporary instance repo.
    '''

    if not machine_name:
        from random import choice
        from string import digits, ascii_lowercase

        nid = ''.join(choice(ascii_lowercase + digits) for _ in range(4))
        machine_name = '{}-{}'.format(osbase.name, nid)

    pi = PackageInjector(osbase)
    if not pi.has_injectables():
        yield pi
    else:
        with temp_dir('pkginject-' + machine_name) as injectrepo_tmp:
            pi.create_instance_repo(injectrepo_tmp)
            yield pi



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import logging as log
from argparse import HelpFormatter, ArgumentParser

from .utils import print_error
from .config import GlobalConfig
from .osbase import OSBase
from .utils.env import set_owning_user, set_unicode_allowed


def init_config(options):
    '''
    Create a new :GlobalConfig from command-line options.
    '''
    gconf = GlobalConfig()
    gconf.load(options.config)

    # check if we are forbidden from using unicode - otherwise we build
    # with unicode enabled by default
    if options.no_unicode:
        set_unicode_allowed(False)
    else:
        import locale

        current_encoding = locale.getpreferredencoding()
        if current_encoding.lower() != 'utf-8':
            log.warning(
                (
                    'Building with unicode support, but your environment does not seem to support unicode. '
                    '(Encoding is {})'
                ).format(current_encoding)
            )
        set_unicode_allowed(True)

    if options.owner:
        info = options.owner.split(':')
        if len(info) > 2:
            print_error('You can only use one colon to split user:group when using the --owner flag.')
            sys.exit(1)
        if len(info) == 2:
            user = info[0]
            group = info[1]
        else:
            user = info[0]
            group = None
        set_owning_user(user, group)

    return gconf


def check_print_version(options):
    if options.show_version:
        from . import __version__

        print(__version__)
        sys.exit(0)


def command_create(options):
    '''Create new container image'''

    check_print_version(options)
    if not options.name:
        print_error('Need at least a container name (suite name) to bootstrap!')
        sys.exit(1)
    gconf = init_config(options)

    components = None
    if options.components:
        components = options.components.split(',')

    extra_suites = []
    if options.extra_suites:
        extra_suites = options.extra_suites.strip().split(' ')

    osbase = OSBase(
        gconf,
        options.suite,
        options.arch,
        variant=options.variant,
        base_suite=options.base_suite,
        custom_name=options.name,
    )
    r = osbase.create(
        options.mirror,
        components,
        extra_suites=extra_suites,
        extra_source_lines=options.extra_source_lines,
        allow_recommends=options.allow_recommends,
        with_init=options.with_init,
    )
    if not r:
        sys.exit(2)


def command_delete(options):
    '''Delete container image'''

    check_print_version(options)
    if not options.name:
        print_error('No container image name was specified!')
        sys.exit(1)
    gconf = init_config(options)
    osbase = OSBase(gconf, options.suite, options.arch, options.variant, custom_name=options.name)
    r = osbase.delete()
    if not r:
        sys.exit(2)


def command_update(options):
    '''Update container image'''

    check_print_version(options)
    if not options.name:
        print_error('Need at least a container image name for update!')
        sys.exit(1)
    gconf = init_config(options)
    osbase = OSBase(gconf, options.suite, options.arch, options.variant, custom_name=options.name)
    if options.recreate:
        r = osbase.recreate()
    else:
        r = osbase.update()
    if not r:
        sys.exit(2)


def command_list(options):
    '''List container images'''

    from .osbase import print_container_base_image_info

    check_print_version(options)
    gconf = init_config(options)
    print_container_base_image_info(gconf)


def command_build(options):
    '''Build a package in a new volatile container'''

    from .build import build_from_dsc, build_from_directory

    check_print_version(options)
    if not options.name:
        print_error('Need at least a container image or suite name for building!')
        sys.exit(1)
    gconf = init_config(options)
    osbase = OSBase(gconf, options.suite, options.arch, options.variant, custom_name=options.name)

    # prepare user-defined environment variables
    env_vars = {}
    if options.env_vars:
        for kv in options.env_vars:
            p = kv.split('=', 1)
            if len(p) != 2:
                print('Environment variable definition `{}` is invalid!'.format(kv))
                print('Can not continue.')
                sys.exit(1)
            env_vars[p[0]] = p[1]

    # prepare user-defined buildflags
    buildflags = []
    if options.buildflags:
        buildflags = options.buildflags.split(';')

    if not options.target and os.path.isdir(options.name):
        print(
            'A directory is given as parameter, but you are missing a container-name parameter to build for.'
        )
        print('Can not continue.')
        sys.exit(1)

    # override globally configured output directory with
    # a custom one defined on the CLI
    if options.results_dir:
        osbase.results_dir = options.results_dir

    if not options.target or os.path.isdir(options.target):
        r = build_from_directory(
            osbase,
            options.target,
            sign=options.sign,
            build_only=options.build_only,
            include_orig=options.include_orig,
            maintainer=options.maintainer,
            clean_source=options.clean_source,
            qa_lintian=options.lintian,
            interact=options.interact,
            log_build=not options.no_buildlog,
            extra_dpkg_flags=buildflags,
            build_env=env_vars,
        )
    else:
        r = build_from_dsc(
            osbase,
            options.target,
            sign=options.sign,
            build_only=options.build_only,
            include_orig=options.include_orig,
            maintainer=options.maintainer,
            qa_lintian=options.lintian,
            interact=options.interact,
            log_build=not options.no_buildlog,
            extra_dpkg_flags=buildflags,
            build_env=env_vars,
        )
    if not r:
        sys.exit(2)


def command_login(options):
    '''Open interactive session in a container'''

    check_print_version(options)
    if not options.name:
        print_error('Need at least a container image name!')
        sys.exit(1)
    gconf = init_config(options)
    osbase = OSBase(gconf, options.suite, options.arch, options.variant, custom_name=options.name)

    allowed = []
    if options.allow:
        allowed = [s.strip() for s in options.allow.split(',')]
    r = osbase.login(options.persistent, allowed=allowed, boot=options.boot)
    if not r:
        sys.exit(2)


def command_run(options, custom_command):
    '''Run arbitrary command in container session'''

    check_print_version(options)
    if not options.name:
        print_error('Need at least a container image name!')
        sys.exit(1)
    gconf = init_config(options)
    osbase = OSBase(
        gconf,
        options.suite,
        options.arch,
        options.variant,
        custom_name=options.name,
        cachekey=options.cachekey,
    )

    allowed = []
    if options.allow:
        allowed = [s.strip() for s in options.allow.split(',')]

    bind_build_dir = options.bind_build_dir
    if bind_build_dir == 'y':
        bind_build_dir = 'ro'
    elif bind_build_dir == 'rw' or bind_build_dir == 'ro':
        pass
    else:
        bind_build_dir = 'n'

    r = osbase.run(
        custom_command,
        options.build_dir,
        options.artifacts_dir,
        boot=options.boot,
        init_command=options.init_command,
        copy_command=options.external_commad,
        header_msg=options.header,
        bind_build_dir=bind_build_dir,
        allowed=allowed,
    )
    if not r:
        sys.exit(2)


def command_maintain(options):
    '''Execute global maintenance actions'''

    check_print_version(options)
    gconf = init_config(options)

    if options.migrate:
        from .maintain import maintain_migrate

        maintain_migrate(gconf)
        return
    if options.clear_caches:
        from .maintain import maintain_clear_caches

        maintain_clear_caches(gconf)
        return
    if options.update_all:
        from .maintain import maintain_update_all

        maintain_update_all(gconf)
        return
    if options.purge:
        from .maintain import maintain_purge

        maintain_purge(gconf, options.yes)
        return
    if options.status:
        from .maintain import maintain_print_status

        maintain_print_status(gconf)
        return

    print_error('No maintenance action selected!')
    sys.exit(1)


class CustomArgparseFormatter(HelpFormatter):
    def _split_lines(self, text, width):
        if text.startswith('CF|'):
            return text[3:].splitlines()
        return HelpFormatter._split_lines(self, text, width)


def add_container_select_arguments(parser):
    parser.add_argument(
        '--variant',
        action='store',
        dest='variant',
        default=None,
        help='Set the bootstrap script variant (use `none` to select no variant).',
    )
    parser.add_argument(
        '-a', '--arch', action='store', dest='arch', default=None, help='The architecture of the container.'
    )
    parser.add_argument(
        '--suite',
        action='store',
        dest='suite',
        default=None,
        help='Explicitly set a suite name (instead of having it derived from the container name).',
    )
    parser.add_argument(
        'name',
        action='store',
        nargs='?',
        default=None,
        help='The name of the container image (usually a distribution suite name).',
    )


def create_parser(formatter_class=None):
    '''Create debspawn CLI argument parser'''

    if not formatter_class:
        formatter_class = CustomArgparseFormatter

    parser = ArgumentParser(description='Build in nspawn containers', formatter_class=formatter_class)
    subparsers = parser.add_subparsers(dest='sp_name', title='subcommands')

    # generic arguments
    parser.add_argument(
        '-c', '--config', action='store', dest='config', default=None, help='Path to the global config file.'
    )
    parser.add_argument('--verbose', action='store_true', dest='verbose', help='Enable debug messages.')
    parser.add_argument(
        '--no-unicode', action='store_true', dest='no_unicode', help='Disable unicode support.'
    )
    parser.add_argument(
        '--version', action='store_true', dest='show_version', help='Display the version of debspawn itself.'
    )

    parser.add_argument(
        '--owner',
        action='store',
        dest='owner',
        default=None,
        help=('Set the user name/uid and group/gid separated by a colon ' 'whose behalf we are acting.'),
    )

    # 'create' command
    sp = subparsers.add_parser('create', help='Create new container image')
    add_container_select_arguments(sp)
    sp.add_argument(
        '--mirror',
        action='store',
        dest='mirror',
        default=None,
        help='Set a specific mirror to bootstrap from.',
    )
    sp.add_argument(
        '--components',
        action='store',
        dest='components',
        default=None,
        help='A comma-separated list of archive components to enable in the newly created image.',
    )
    sp.add_argument(
        '--base-suite',
        action='store',
        dest='base_suite',
        default=None,
        help=(
            'A full suite that forms the base of the selected partial suite '
            '(e.g. for -updates and -backports).'
        ),
    )
    sp.add_argument(
        '--extra-suites',
        action='store',
        dest='extra_suites',
        default=None,
        help=(
            'Space-separated list of additional suites that should also be added to the ' 'sources.list file.'
        ),
    )
    sp.add_argument(
        '--extra-sourceslist-lines',
        action='store',
        dest='extra_source_lines',
        default=None,
        help=(
            'Lines that should be added to the build environments source.list verbatim. '
            'Separate lines by linebreaks.'
        ),
    )
    sp.add_argument(
        '--allow-recommends',
        action='store',
        dest='allow_recommends',
        default=None,
        help=(
            'Do not disable APT installing "recommends"-type dependencies by default, and instead use '
            'the default behavior for full, normal system installations with "recommends" enabled.'
        ),
    )
    sp.add_argument(
        '--with-init',
        action='store_true',
        dest='with_init',
        help='Include an init system in this image, so it is bootable.',
    )
    sp.set_defaults(func=command_create)

    # 'delete' command
    sp = subparsers.add_parser('delete', help='Remove a container image')
    add_container_select_arguments(sp)
    sp.set_defaults(func=command_delete)

    # 'update' command
    sp = subparsers.add_parser('update', help='Update a container image')
    add_container_select_arguments(sp)
    sp.add_argument(
        '--recreate',
        action='store_true',
        dest='recreate',
        help=(
            'Re-create the container image from scratch using the settings used to create it previously, '
            'instead of just updating it.'
        ),
    )
    sp.set_defaults(func=command_update)

    # 'list' command
    sp = subparsers.add_parser('list', help='List available container images', aliases=['ls'])
    sp.set_defaults(func=command_list)

    # 'build' command
    sp = subparsers.add_parser(
        'build',
        help='Build a package in an isolated environment',
        formatter_class=formatter_class,
        aliases=['b'],
    )
    add_container_select_arguments(sp)
    sp.add_argument('--sign', action='store_true', dest='sign', help='Sign the resulting package.')
    sp.add_argument(
        '--only',
        choices=['binary', 'arch', 'indep', 'source'],
        dest='build_only',
        help=(
            'CF|Select only a specific set of packages to be built. Choices are:\n'
            'binary: Build only binary packages, no source files are to be built and/or distributed.\n'
            'arch: Build only architecture-specific binary packages.\n'
            'indep: Build only architecture-independent (arch:all) binary packages.\n'
            'source: Do a source-only build, no binary packages are made.'
        ),
    )
    sp.add_argument(
        '--include-orig',
        action='store_true',
        dest='include_orig',
        help='Forces the inclusion of the original source.',
    )
    sp.add_argument(
        '--buildflags',
        action='store',
        dest='buildflags',
        help='Set flags passed through to dpkg-buildpackage as semicolon-separated list.',
    )
    sp.add_argument(
        '--results-dir',
        action='store',
        dest='results_dir',
        help='Override the configured results directory and return artifacts at a custom location.',
    )
    sp.add_argument(
        '--maintainer',
        action='store',
        dest='maintainer',
        help=(
            'Set the name and email address of the maintainer for this package and upload, rather than '
            'using the information from the source tree\'s control file or changelog.'
        ),
    )
    sp.add_argument(
        '--clean-source',
        action='store_true',
        dest='clean_source',
        help=(
            'Run the d/rules clean target outside of the container. This means the package build '
            'dependencies need to be installed on the host system when building from a '
            'source directory.'
        ),
    )
    sp.add_argument(
        '--lintian',
        action='store_true',
        dest='lintian',
        help='Run the Lintian static analysis tool for Debian packages after the package is built.',
    )
    sp.add_argument(
        '--no-buildlog', action='store_true', dest='no_buildlog', help='Do not write a build log.'
    )
    sp.add_argument(
        '-i',
        '--interact',
        action='store_true',
        dest='interact',
        help=(
            'Run an interactive shell in the build environment after build. This implies `--no-buildlog` '
            'and disables the log.'
        ),
    )
    sp.add_argument(
        '-e',
        '--setenv',
        action='append',
        dest='env_vars',
        help=(
            'Set an environment variable for the build process. Takes a `key=value` pair any may be '
            'defined multiple times to set different environment variables.'
        ),
    )
    sp.add_argument(
        'target',
        action='store',
        nargs='?',
        default=None,
        help='The source package file or source directory to build.',
    )
    sp.set_defaults(func=command_build)

    # 'login' command
    sp = subparsers.add_parser('login', help='Open interactive session in a container')
    add_container_select_arguments(sp)
    sp.add_argument(
        '--persistent',
        action='store_true',
        dest='persistent',
        help='Make changes done in the session persistent.',
    )
    sp.add_argument(
        '--allow',
        action='store',
        dest='allow',
        help=(
            'List one or more additional permissions to grant the container. Takes a comma-separated '
            'list of capability names.'
        ),
    )
    sp.add_argument(
        '--boot',
        action='store_true',
        dest='boot',
        help='Boot container image (requires the image to contain an init system).',
    )
    sp.set_defaults(func=command_login)

    # 'run' command
    sp = subparsers.add_parser('run', help='Run arbitrary command in an ephemeral container')
    add_container_select_arguments(sp)
    sp.add_argument(
        '--artifacts-out',
        action='store',
        dest='artifacts_dir',
        default=None,
        help='Directory on the host where artifacts can be stored. Mounted to /srv/artifacts in the guest.',
    )
    sp.add_argument(
        '--build-dir',
        action='store',
        dest='build_dir',
        default=None,
        help='Select a host directory that gets copied to /srv/build.',
    )
    sp.add_argument(
        '--bind-build-dir',
        action='store',
        dest='bind_build_dir',
        default='n',
        choices=['y', 'n', 'ro', 'rw'],
        help=(
            'Bindmount build directory instead of copying it. Mounts read-only by default, but can mount '
            'as writable as well if \'rw\' is passed as value.'
        ),
    )
    sp.add_argument(
        '--cachekey',
        action='store',
        dest='cachekey',
        default=None,
        help=(
            'If set, use the specified cache-ID to store an initialized container image for faster '
            'initialization times.\n'
            'This may mean that the command passed in `--init-command` is skipped if the cache '
            'already existed.'
        ),
    )
    sp.add_argument(
        '--init-command',
        action='store',
        dest='init_command',
        default=None,
        help='The command or command script used to set up the container.',
    )
    sp.add_argument(
        '-x',
        '--external-command',
        action='store_true',
        dest='external_commad',
        help=(
            'If set, the command script(s) will be copied from the host to the container '
            'and then executed.'
        ),
    )
    sp.add_argument(
        '--header',
        action='store',
        dest='header',
        default=None,
        help='Name of the task that is run, will be printed as header.',
    )
    sp.add_argument(
        '--allow',
        action='store',
        dest='allow',
        help=(
            'List one or more additional permissions to grant the container. Takes a comma-separated '
            'list of capability names.'
        ),
    )
    sp.add_argument(
        '--boot',
        action='store_true',
        dest='boot',
        help='Boot container image (requires the image to contain an init system).',
    )
    sp.add_argument('command', action='store', nargs='*', default=None, help='The command to run.')

    # 'maintain' command
    sp = subparsers.add_parser('maintain', help='Execute various maintenance actions, affecting all images')
    sp.add_argument(
        '-y', '--yes', action='store_true', dest='yes', help='Perform dangerous actions without asking twice.'
    )
    sp.add_argument(
        '--migrate',
        action='store_true',
        dest='migrate',
        help='Migrate any settings or configuration changes to the current version of debspawn.',
    )
    sp.add_argument(
        '--update-all',
        action='store_true',
        dest='update_all',
        help='Update all container images that we know.',
    )
    sp.add_argument(
        '--clear-caches',
        action='store_true',
        dest='clear_caches',
        help='Delete all cached packages for all images.',
    )
    sp.add_argument(
        '--purge',
        action='store_true',
        dest='purge',
        help='Remove all images as well as any data associated with them.',
    )
    sp.add_argument(
        '--status',
        action='store_true',
        dest='status',
        help='Display a status summary about this installation, highlighting potential issues.',
    )
    sp.set_defaults(func=command_maintain)

    return parser


def run(mainfile, args):
    if len(args) == 0:
        print_error('Need a subcommand to proceed!')
        sys.exit(1)

    parser = create_parser()

    # special case, so 'run' can understand which arguments are for debspawn and which are
    # for the command to be executed
    custom_command = None
    if args[0] == 'run':
        for i, arg in enumerate(args):
            if arg == '---':
                if i + 1 == len(args):
                    print_error('No command was given after "---", can not continue.')
                    sys.exit(1)
                custom_command = args[i + 1 :]
                args = args[:i]
                break

    args = parser.parse_args(args)
    check_print_version(args)
    if args.sp_name == 'run':
        if not custom_command:
            custom_command = args.command
        command_run(args, custom_command)
    else:
        if not hasattr(args, 'func'):
            print_error('Unknown or no subcommand was provided. Can not proceed.')
            sys.exit(1)
        args.func(args)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

__appname__ = 'debspawn'
__version__ = '0.6.2'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import shutil
from contextlib import contextmanager

_unicode_allowed = True  # store whether we are allowed to use unicode
_owner_uid = 0  # uid of the user on whose behalf we are running
_owner_gid = 0  # gid of the user on whose behalf we are running


def set_owning_user(user, group=None):
    '''
    Set the user on whose behalf we are running.
    This is useful so we can drop privileges to
    the perticular user in many cases.
    '''
    from grp import getgrnam
    from pwd import getpwnam, getpwuid

    if user.isdecimal():
        uid = int(user)
    else:
        uid = getpwnam(user).pw_uid

    if not group:
        gid = getpwuid(uid).pw_gid
    elif group.isdecimal():
        gid = int(group)
    else:
        gid = getgrnam(group).gr_gid

    global _owner_uid
    global _owner_gid
    _owner_uid = uid
    _owner_gid = gid


def ensure_root():
    '''
    Ensure we are running as root and all code following
    this function is privileged.
    '''
    if os.geteuid() == 0:
        return

    args = sys.argv.copy()

    owner_set = any(a.startswith('--owner=') for a in sys.argv)
    if owner_set:
        # we don't override an owner explicitly set by the user
        args = sys.argv.copy()
    else:
        args = [sys.argv[0]]

        # set flag to tell the new process who it can impersonate
        # for unprivileged actions. It it is root, just omit the flag.
        uid = os.getuid()
        gid = os.getgid()
        if uid != 0 or gid != 0:
            args.append('--owner={}:{}'.format(uid, gid))
        args.extend(sys.argv[1:])

    def filter_env_far(result, name):
        value = os.environ.get(name)
        if not value:
            return
        result.append('{}={}'.format(name, shlex.quote(value)))

    if shutil.which('sudo'):
        # Filter "good" environment variables that we want to have after running sudo.
        # Most of those are standard variables affecting debsign bahevior later, in case
        # the user has requested signing
        import shlex

        env = []
        filter_env_far(env, 'DEBEMAIL')
        filter_env_far(env, 'DEBFULLNAME')
        filter_env_far(env, 'GPGKEY')
        filter_env_far(env, 'GPG_AGENT_INFO')
        filter_env_far(env, 'HTTP_PROXY')
        filter_env_far(env, 'HTTPS_PROXY')
        filter_env_far(env, 'http_proxy')
        filter_env_far(env, 'https_proxy')

        os.execvp("sudo", ["sudo"] + env + args)
    else:
        print('This command needs to be run as root.')
        sys.exit(1)


@contextmanager
def switch_unprivileged():
    '''
    Run actions using the unprivileged user ID
    on the behalf of which we are running.
    This is NOT a security feature!
    '''
    import pwd

    if _owner_uid == 0 and _owner_gid == 0:
        # we can't really do much here, we have to run
        # as root, as we don't know an unprivileged user
        # to switch to

        yield
    else:
        orig_egid = os.getegid()
        orig_euid = os.geteuid()
        orig_home = os.environ.get('HOME')
        if not orig_home:
            orig_home = pwd.getpwuid(os.getuid()).pw_dir

        try:
            os.setegid(_owner_gid)
            os.seteuid(_owner_uid)
            os.environ['HOME'] = pwd.getpwuid(_owner_uid).pw_dir

            yield
        finally:
            os.setegid(orig_egid)
            os.seteuid(orig_euid)
            os.environ['HOME'] = orig_home


def get_owner_uid_gid():
    return _owner_uid, _owner_gid


def get_random_free_uid_gid():
    '''Get a random unused UID and GID for the current system.'''
    import pwd
    import random

    uid = 1000
    gid = 1000
    for pw in pwd.getpwall():
        if pw.pw_uid > uid:
            uid = pw.pw_uid
        if pw.pw_gid > gid:
            gid = pw.pw_gid
    # we can not use an extremely large number here, as otherwise the container's
    # lastlog/faillog will grow to insane sizes
    r = random.randint(100, 2048)
    return uid + r, gid + r


def colored_output_allowed():
    return (hasattr(sys.stdout, "isatty") and sys.stdout.isatty()) or (
        'TERM' in os.environ and os.environ['TERM'] == 'ANSI'
    )


def unicode_allowed():
    return _unicode_allowed


def set_unicode_allowed(val):
    global _unicode_allowed
    _unicode_allowed = val


def get_free_space(path):
    '''
    Return free space of :path
    '''
    real_path = os.path.realpath(path)
    stat = os.statvfs(real_path)
    # get free space in MiB.
    free_space = float(stat.f_bsize * stat.f_bavail)
    return free_space


def get_tree_size(path):
    '''
    Return total size of files in path and subdirs. If
    is_dir() or stat() fails, print an error message to stderr
    and assume zero size (for example, file has been deleted).
    '''
    total = 0
    for entry in os.scandir(path):
        try:
            is_dir = entry.is_dir(follow_symlinks=False)
        except OSError as error:
            print('Error calling is_dir():', error, file=sys.stderr)
            continue
        if is_dir:
            total += get_tree_size(entry.path)
        else:
            try:
                total += entry.stat(follow_symlinks=False).st_size
            except OSError as error:
                print('Error calling stat():', error, file=sys.stderr)
    return total



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import shutil

from .command import run_command


class TarError(Exception):
    """Generic error while using tar/zstd."""


def ensure_tar_zstd():
    '''Check if the required binaries for compression are available'''

    if not shutil.which('zstd'):
        raise TarError(
            (
                'The "zsdt" binary was not found, we can not compress tarballs. '
                'Please install zstd to continue!'
            )
        )
    if not shutil.which('tar'):
        raise TarError(
            'The "tar" binary was not found, we can not create tarballs. Please install tar to continue!'
        )


def compress_directory(dirname, tarname):
    '''Compress a directory to a given tarball'''

    cmd = ['tar', '-C', dirname, '-I', 'zstd', '-cf', tarname, '.']

    out, err, ret = run_command(cmd)

    if ret != 0:
        raise TarError('Unable to create tarball "{}":\n{}{}'.format(tarname, out, err))


def decompress_tarball(tarname, dirname):
    '''Compress a directory to a given tarball'''

    cmd = ['tar', '-C', dirname, '-I', 'zstd', '-xf', tarname]

    out, err, ret = run_command(cmd)

    if ret != 0:
        raise TarError('Unable to decompress tarball "{}":\n{}{}'.format(tarname, out, err))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import json
import shutil
from glob import glob

from .config import GlobalConfig
from .osbase import OSBase
from .utils.env import ensure_root
from .utils.log import (
    print_info,
    print_warn,
    print_error,
    print_bullet,
    print_section,
    print_bool_item,
)


def ensure_rmtree_symattack_protection():
    '''Exit the program with an error if rmtree does not protect against symlink attacks'''
    if not shutil.rmtree.avoids_symlink_attacks:
        print_error('Will not continue: rmtree does not run in symlink-attack protected mode.')
        sys.exit(1)


def maintain_migrate(gconf: GlobalConfig):
    '''Migrate configuration from older versions of debspawn to the latest version'''

    ensure_root()

    # migrate old container images directory, if needed
    images_dir = gconf.osroots_dir
    if not os.path.isdir(images_dir):
        old_images_dir = '/var/lib/debspawn/containers'
        if os.path.isdir(old_images_dir):
            print_info('Migrating images directory...')
            shutil.move(old_images_dir, images_dir)


def maintain_clear_caches(gconf: GlobalConfig):
    '''Delete all cache data for all images'''

    ensure_root()
    ensure_rmtree_symattack_protection()

    aptcache_dir = gconf.aptcache_dir
    if os.path.isdir(aptcache_dir):
        for cdir in glob(os.path.join(aptcache_dir, '*')):
            print_info('Removing APT cache for: {}'.format(os.path.basename(cdir)))
            if os.path.isdir(cdir):
                shutil.rmtree(cdir)
            else:
                os.remove(cdir)

    dcache_dir = os.path.join(gconf.osroots_dir, 'dcache')
    if os.path.isdir(dcache_dir):
        print_info('Removing image derivatives cache.')
        shutil.rmtree(dcache_dir)


def maintain_purge(gconf: GlobalConfig, force: bool = False):
    '''Remove all images as well as any data associated with them'''

    ensure_root()
    ensure_rmtree_symattack_protection()

    if not force:
        print_warn(
            (
                'This action will delete ALL your images as well as their configuration, build results and other '
                'associated data and will clear all data from the directories you may have configured as default.'
            )
        )
        delete_all = False
        while True:
            try:
                in_res = input('Do you really want to continue? [y/N]: ')
            except EOFError:
                in_res = 'n'
                print()
            if not in_res:
                delete_all = False
                break
            elif in_res.lower() == 'y':
                delete_all = True
                break
            elif in_res.lower() == 'n':
                delete_all = False
                break

        if not delete_all:
            print_info('Purge action aborted.')
            return

    print_warn('Deleting all images, image configuration, build results and state data.')
    for sdir in [gconf.osroots_dir, gconf.results_dir, gconf.aptcache_dir, gconf.injected_pkgs_dir]:
        if not os.path.isdir(sdir):
            continue
        if sdir.startswith('/home/') or sdir.startswith('/usr/'):
            continue
        print_info('Purging: {}'.format(sdir))
        for d in glob(os.path.join(sdir, '*')):
            if os.path.isdir(d):
                shutil.rmtree(d)
            else:
                os.remove(d)

    default_state_dir = '/var/lib/debspawn/'
    if os.path.isdir(default_state_dir):
        print_info('Removing: {}'.format(default_state_dir))
        shutil.rmtree(default_state_dir)


def maintain_update_all(gconf: GlobalConfig):
    '''Update all container images that we know.'''

    ensure_root()

    osroots_dir = gconf.osroots_dir
    tar_files = []
    if os.path.isdir(osroots_dir):
        tar_files = list(glob(os.path.join(osroots_dir, '*.tar.zst')))
    if not tar_files:
        print_info('No container base images have been found!')
        return

    failed_images = []
    nodata_images = []
    first_entry = True
    for tar_fname in tar_files:
        img_basepath = os.path.splitext(os.path.splitext(tar_fname)[0])[0]
        config_fname = img_basepath + '.json'
        imgid = os.path.basename(img_basepath)

        # read configuration data
        if not os.path.isfile(config_fname):
            nodata_images.append(imgid)
            continue

        with open(config_fname, 'rt') as f:
            cdata = json.loads(f.read())

        if not first_entry:
            print()
        first_entry = False
        print_bullet('Update: {}'.format(imgid), indent=1, large=True)

        osbase = OSBase(
            gconf,
            cdata['Suite'],
            cdata['Architecture'],
            cdata.get('Variant'),
            custom_name=os.path.basename(img_basepath),
        )
        r = osbase.update()
        if not r:
            print_error('Failed to update {}'.format(imgid))
            failed_images.append(imgid)

    if nodata_images or failed_images:
        print()
    for imgid in nodata_images:
        print_warn('Could not auto-update image {}: Configuration data is missing.'.format(imgid))
    if failed_images:
        print_error('Failed to update image(s): {}'.format(', '.join(failed_images)))
        sys.exit(1)


def maintain_print_status(gconf: GlobalConfig):
    '''
    Print status information about this Debspawn installation
    that may be useful for debugging issues.
    '''
    import platform

    from . import __version__
    from .nspawn import systemd_version, systemd_detect_virt
    from .osbase import bootstrap_tool_version, print_container_base_image_info

    print('Debspawn Status Report', end='')
    sys.stdout.flush()

    # read distribution information
    os_release = {}
    if os.path.exists('/etc/os-release'):
        with open('/etc/os-release') as f:
            for line in f:
                k, v = line.rstrip().split("=")
                os_release[k] = v.strip('"')

    print_section('Host System')
    print('OS:', os_release.get('NAME', 'Unknown'), os_release.get('VERSION', '<?>'))
    print('Platform:', platform.platform(aliased=True))
    print('Virtualization:', systemd_detect_virt())
    print('Systemd-nspawn version:', systemd_version())
    print('Bootstrap tool:', '{} {}'.format(gconf.bootstrap_tool, bootstrap_tool_version(gconf)))

    print_section('Container image list')
    print_container_base_image_info(gconf)

    print_section('Debspawn')
    print('Version:', __version__)
    print_bool_item(
        'Tmpfiles.d configuration:',
        os.path.isfile('/usr/lib/tmpfiles.d/debspawn.conf'),
        text_true='installed',
        text_false='missing',
    )
    print_bool_item(
        'Monthly cache cleanup timer:',
        os.path.isfile('/lib/systemd/system/debspawn-clear-caches.timer'),
        text_true='available',
        text_false='missing',
    )
    print_bool_item(
        'Manual pages:',
        len(glob('/usr/share/man/man1/debspawn*.1.*')) >= 8,
        text_true='installed',
        text_false='missing',
    )
    if not os.path.isfile('/etc/debspawn/global.toml'):
        print('Global configuration: default')
    else:
        print('Global configuration:')
        with open('/etc/debspawn/global.toml', 'r') as f:
            for line in f:
                print('    ', line)



# -*- coding: utf-8 -*-
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from .env import unicode_allowed, colored_output_allowed
from .log import print_info, print_warn, print_error, print_header, print_section
from .misc import (
    cd,
    listify,
    temp_dir,
    rmtree_mntsafe,
    systemd_escape,
    format_filesize,
    hardlink_or_copy,
)
from .command import safe_run, run_forwarded

__all__ = [
    'print_info',
    'print_warn',
    'print_error',
    'print_header',
    'print_section',
    'colored_output_allowed',
    'unicode_allowed',
    'safe_run',
    'run_forwarded',
    'listify',
    'temp_dir',
    'cd',
    'hardlink_or_copy',
    'format_filesize',
    'rmtree_mntsafe',
    'systemd_escape',
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import stat
import fcntl
import shutil
import typing as T
import subprocess
from pathlib import Path
from contextlib import contextmanager

from ..config import GlobalConfig


class MountError(Exception):
    """Error while dealing with mountpoints."""


def listify(item: T.Any):
    '''
    Return a list of :item, unless :item already is a lit.
    '''
    if not item:
        return []
    return item if type(item) == list else [item]


@contextmanager
def cd(where):
    ncwd = os.getcwd()
    try:
        yield os.chdir(where)
    finally:
        os.chdir(ncwd)


def random_string(prefix: T.Optional[str] = None, count: int = 8):
    '''
    Create a string of random alphanumeric characters of a given length,
    separated with a hyphen from an optional prefix.
    '''

    from random import choice
    from string import digits, ascii_lowercase

    if count <= 0:
        count = 1
    rdm_id = ''.join(choice(ascii_lowercase + digits) for _ in range(count))
    if prefix:
        return '{}-{}'.format(prefix, rdm_id)
    return rdm_id


def systemd_escape(name: str) -> T.Optional[str]:
    '''Escape a string using systemd's escaping rules.'''
    from .command import run_command

    out, _, ret = run_command(['systemd-escape', name])
    if ret != 0:
        return None
    return out.strip()


@contextmanager
def temp_dir(basename=None):
    '''Context manager for a temporary directory in debspawn's temp-dir location.

    This function will also ensure that we will not jump into possibly still
    bind-mounted directories upon deletion, and will unmount those directories
    instead.
    '''

    dir_name = random_string(basename)
    temp_basedir = GlobalConfig().temp_dir
    if not temp_basedir:
        temp_basedir = '/var/tmp/debspawn/'

    tmp_path = os.path.join(temp_basedir, dir_name)
    Path(tmp_path).mkdir(parents=True, exist_ok=True)

    fd = os.open(tmp_path, os.O_RDONLY)
    # we hold a shared lock on the directory to prevent systemd-tmpfiles
    # from deleting it, just in case we are building something for days
    try:
        if fd > 0:
            fcntl.flock(fd, fcntl.LOCK_SH | fcntl.LOCK_NB)
    except (IOError, OSError):
        print('WARNING: Unable to lock temporary directory {}'.format(tmp_path), file=sys.stderr)
        sys.stderr.flush()

    try:
        yield tmp_path
    finally:
        try:
            fcntl.flock(fd, fcntl.LOCK_UN)
            rmtree_mntsafe(tmp_path)
        finally:
            if fd > 0:
                os.close(fd)


def safe_copy(src, dst, *, preserve_mtime: bool = True):
    '''
    Attempt to safely copy a file, by atomically replacing the destination and
    protecting against symlink attacks.
    '''
    dst_tmp = random_string(dst + '.tmp')
    try:
        if preserve_mtime:
            shutil.copy2(src, dst_tmp)
        else:
            shutil.copy(src, dst_tmp)
        if os.path.islink(dst):
            os.remove(dst)
        os.replace(dst_tmp, dst)
    finally:
        try:
            os.remove(dst_tmp)
        except OSError:
            pass


def maybe_remove(f):
    '''Delete a file if it exists, but do nothing if it doesn't.'''
    try:
        os.remove(f)
    except OSError:
        pass


def format_filesize(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def current_time_string():
    '''Get the current time as human-readable string.'''

    from datetime import datetime, timezone

    utc_dt = datetime.now(timezone.utc)
    return utc_dt.astimezone().strftime('%Y-%m-%d %H:%M:%S UTC%z')


def version_noepoch(version):
    '''Return version from :version without epoch.'''

    version_noe = version
    if ':' in version_noe:
        version_noe = version_noe.split(':', 1)[1]
    return version_noe


def hardlink_or_copy(src, dst):
    '''Hardlink a file :src to :dst or copy the file in case linking is not possible'''

    try:
        os.link(src, dst)
    except (PermissionError, OSError):
        shutil.copy2(src, dst)


def is_mountpoint(path) -> bool:
    '''Check if :path is a mountpoint.

    Unlike os.path.ismount, this function will also consider
    bindmountpoints.
    This function may be slow
    '''

    if not os.path.exists(path):
        return False
    if os.path.ismount(path):
        return True

    ret = subprocess.run(['findmnt', '-M', str(path)], capture_output=True, check=False)
    if ret.returncode == 0:
        return True
    return False


def bindmount(from_path, to_path):
    '''Create a bindmount point.'''

    cmd = ['mount', '--bind', from_path, to_path]
    ret = subprocess.run(cmd, capture_output=True, check=False)
    if ret.returncode != 0:
        raise MountError('Unable to create bindmount {} -> {}'.format(from_path, to_path))


def umount(path, lazy: bool = True):
    '''Try to unmount a path.'''

    cmd = ['umount']
    if lazy:
        cmd.append('-l')
    cmd.append(path)
    ret = subprocess.run(cmd, capture_output=True, check=False)
    if ret.returncode != 0:
        raise MountError('Unable to umount path {}'.format(path))

    # try again if the mountpoint is still there, as
    # overmounting may have happened
    if is_mountpoint(path):
        umount(path, lazy=lazy)


def _rmtree_mntsafe_fd(topfd, path, onerror):
    try:
        with os.scandir(topfd) as scandir_it:
            entries = list(scandir_it)
    except OSError as err:
        err.filename = path
        onerror(os.scandir, path, sys.exc_info())
        return
    for entry in entries:
        fullname = os.path.join(path, entry.name)
        try:
            is_dir = entry.is_dir(follow_symlinks=False)
        except OSError:
            is_dir = False
        else:
            if is_dir:
                try:
                    orig_st = entry.stat(follow_symlinks=False)
                    is_dir = stat.S_ISDIR(orig_st.st_mode)
                except OSError:
                    onerror(os.lstat, fullname, sys.exc_info())
                    continue
        if is_dir:
            if is_mountpoint(fullname):
                try:
                    umount(fullname)
                    orig_st = os.stat(fullname, follow_symlinks=False)
                except Exception:
                    onerror(umount, fullname, sys.exc_info())
                    continue

            try:
                dirfd = os.open(entry.name, os.O_RDONLY, dir_fd=topfd)
            except OSError:
                onerror(os.open, fullname, sys.exc_info())
            else:
                try:
                    if os.path.samestat(orig_st, os.fstat(dirfd)):
                        _rmtree_mntsafe_fd(dirfd, fullname, onerror)
                        try:
                            os.rmdir(entry.name, dir_fd=topfd)
                        except OSError:
                            onerror(os.rmdir, fullname, sys.exc_info())
                    else:
                        try:
                            # This can only happen if someone replaces
                            # a directory with a symlink after the call to
                            # os.scandir or stat.S_ISDIR above.
                            raise OSError('Cannot call rmtree on a symbolic link')
                        except OSError:
                            onerror(os.path.islink, fullname, sys.exc_info())
                finally:
                    os.close(dirfd)
        else:
            try:
                os.unlink(entry.name, dir_fd=topfd)
            except OSError:
                onerror(os.unlink, fullname, sys.exc_info())


def rmtree_mntsafe(path, ignore_errors=False, onerror=None):
    '''Recursively delete a directory tree, unmounting mount points if possible.
    This function is based on shutil.rmtree, but will not jump into mount points,
    but instead try to unmount them and if that fails leave them alone.
    This prevents data loss in case bindmounts were set carelessly.

    If ignore_errors is set, errors are ignored; otherwise, if onerror
    is set, it is called to handle the error with arguments (func,
    path, exc_info) where func is platform and implementation dependent;
    path is the argument to that function that caused it to fail; and
    exc_info is a tuple returned by sys.exc_info().  If ignore_errors
    is false and onerror is None, an exception is raised.
    '''
    if ignore_errors:
        # pylint: disable=function-redefined
        def onerror(*args):
            pass

    elif onerror is None:
        # pylint: disable=misplaced-bare-raise
        def onerror(*args):
            raise

    # While the unsafe rmtree works fine on bytes, the fd based does not.
    if isinstance(path, bytes):
        path = os.fsdecode(path)

    if os.path.ismount(path):
        try:
            umount(path)
        except Exception:
            onerror(umount, path, sys.exc_info())
            return

    # Note: To guard against symlink races, we use the standard
    # lstat()/open()/fstat() trick.
    try:
        orig_st = os.lstat(path)
    except Exception:
        onerror(os.lstat, path, sys.exc_info())
        return
    try:
        fd = os.open(path, os.O_RDONLY)
    except Exception:
        onerror(os.open, path, sys.exc_info())
        return
    try:
        if os.path.samestat(orig_st, os.fstat(fd)):
            _rmtree_mntsafe_fd(fd, path, onerror)
            try:
                os.rmdir(path)
            except OSError:
                onerror(os.rmdir, path, sys.exc_info())
        else:
            try:
                # symlinks to directories are forbidden, see bug #1669
                raise OSError("Cannot call rmtree on a symbolic link")
            except OSError:
                onerror(os.path.islink, path, sys.exc_info())
    finally:
        os.close(fd)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2022 Matthias Klumpp <matthias@tenstral.net>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

#
# This is a helper script to install additional configuration and documentation into
# system locations, which Python's setuptools and pip will not usually let us install.
#

import os
import sys
import shutil
from pathlib import Path
from argparse import ArgumentParser
from tempfile import TemporaryDirectory

try:
    import pkgconfig
except ImportError:
    print()
    print(
        (
            'Unable to import pkgconfig. Please install the module '
            '(apt install python3-pkgconfig or pip install pkgconfig) '
            'to continue.'
        )
    )
    print()
    sys.exit(4)
from docs.assemble_man import create_manpage, generate_docbook_pages


class Installer:
    def __init__(self, root: str = None, prefix: str = None):
        if not root:
            root = os.environ.get('DESTDIR')
        if not root:
            root = '/'
        self.root = root

        if not prefix:
            prefix = '/usr/local' if self.root == '/' else '/usr'
        if prefix.startswith('/'):
            prefix = prefix[1:]
        self.prefix = prefix

    def install(self, src, dst, replace_vars=False):
        if dst.startswith('/'):
            dst = dst[1:]
            dst_full = os.path.join(self.root, dst, os.path.basename(src))
        else:
            dst_full = os.path.join(self.root, self.prefix, dst, os.path.basename(src))

        Path(os.path.dirname(dst_full)).mkdir(mode=0o755, parents=True, exist_ok=True)
        if replace_vars:
            with open(src, 'r') as f_src:
                with open(dst_full, 'w') as f_dst:
                    for line in f_src:
                        f_dst.write(line.replace('@PREFIX@', '/' + self.prefix))
        else:
            shutil.copy(src, dst_full)
        os.chmod(dst_full, 0o644)
        print('{}\t\t{}'.format(os.path.basename(src), dst_full))


def chdir_to_source_root():
    thisfile = __file__
    if not os.path.isabs(thisfile):
        thisfile = os.path.normpath(os.path.join(os.getcwd(), thisfile))
    os.chdir(os.path.dirname(thisfile))


def make_manpages(temp_dir):
    '''Build manual pages'''

    # check for xsltproc, we need it to build manual pages
    if not shutil.which('xsltproc'):
        print('The "xsltproc" binary was not found. Please install it to continue!')
        sys.exit(1)

    build_dir = os.path.join(temp_dir, 'docbook')
    Path(build_dir).mkdir(parents=True, exist_ok=True)
    pages = generate_docbook_pages(build_dir)
    man_files = []
    for page in pages:
        man_files.append(create_manpage(page, temp_dir))
    return man_files


def install_data(temp_dir: str, root_dir: str, prefix_dir: str):
    chdir_to_source_root()

    print('Checking dependencies')
    if not pkgconfig.installed('systemd', '>= 240'):
        print('Systemd is not installed on this system. Please make systemd available to continue.')
        sys.exit(4)

    print('Generating manual pages')
    manpage_files = make_manpages(temp_dir)

    print('Installing data')
    inst = Installer(root_dir, prefix_dir)
    sd_tmpfiles_dir = pkgconfig.variables('systemd')['tmpfilesdir']
    sd_system_unit_dir = pkgconfig.variables('systemd')['systemdsystemunitdir']
    man_dir = os.path.join('share', 'man', 'man1')

    inst.install('data/tmpfiles.d/debspawn.conf', sd_tmpfiles_dir)
    inst.install('data/services/debspawn-clear-caches.timer', sd_system_unit_dir)
    inst.install('data/services/debspawn-clear-caches.service', sd_system_unit_dir, replace_vars=True)
    for mf in manpage_files:
        inst.install(mf, man_dir)


def main():
    parser = ArgumentParser(description='Debspawn system data installer')

    parser.add_argument(
        '--root', action='store', dest='root', default=None, help='Root directory to install into.'
    )
    parser.add_argument(
        '--prefix',
        action='store',
        dest='prefix',
        default=None,
        help='Directory prefix (usually `/usr` or `/usr/local`).',
    )

    options = parser.parse_args(sys.argv[1:])
    with TemporaryDirectory(prefix='dsinstall-') as temp_dir:
        install_data(temp_dir, options.root, options.prefix)
    return 0


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2022 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import json
import shutil
import typing as T
import subprocess
from pathlib import Path
from contextlib import contextmanager

from .utils import (
    listify,
    temp_dir,
    print_info,
    print_warn,
    print_error,
    print_header,
    print_section,
    format_filesize,
)
from .config import GlobalConfig
from .nspawn import nspawn_run_persist, nspawn_run_helper_persist
from .aptcache import APTCache
from .utils.env import ensure_root, get_owner_uid_gid, get_random_free_uid_gid
from .utils.misc import safe_copy, maybe_remove, rmtree_mntsafe
from .utils.command import safe_run
from .utils.zstd_tar import ensure_tar_zstd, compress_directory, decompress_tarball


def bootstrap_tool_version(gconf=None):
    if not gconf:
        gconf = GlobalConfig()
    ds_version = 'unknown'
    try:
        out, _, _ = safe_run([gconf.bootstrap_tool, '--version'])
        parts = out.strip().split(' ', 2)
        ds_version = parts[0 if len(parts) < 2 else 1]
    except Exception as e:
        print_warn('Unable to determine bootstrap tool version: {}'.format(e))

    return ds_version


class OSBase:
    '''
    Describes an OS base registered with debspawn
    '''

    def __init__(
        self,
        gconf: GlobalConfig,
        suite: str,
        arch: str,
        variant=None,
        *,
        base_suite=None,
        custom_name=None,
        cachekey=None,
    ):
        self._gconf = gconf
        self._suite = suite
        self._base_suite = base_suite
        self._arch = arch

        # if variant is not passed, we use the configured default
        self._variant = variant if variant else gconf.default_bootstrap_variant
        if self._variant == 'none':
            # "none" is an alias to "don't set a variant when invoking debootstrap"
            self._variant = None

        self._custom_name = custom_name
        self._name = self._make_name()
        self._results_dir = self._gconf.results_dir
        self._cachekey = cachekey
        if self._cachekey:
            self._cachekey = self._cachekey.replace(' ', '')

        self._parameters_checked = False
        self._aptcache = APTCache(self)

        # debootstrap-compatible tools that we know about
        self._known_bootstrap_tools = ('debootstrap', 'mmdebstrap', 'qemu-debootstrap')

        # get a fresh UID to give to our build user within the container
        self._builder_uid = get_random_free_uid_gid()[0]

        # ensure we can (de)compress zstd tarballs
        ensure_tar_zstd()

    def _make_name(self):
        '''Configure a unique-ish name based on user defined data,
        and tweak the custom name and suite values to match.'''

        if self._custom_name and not self._suite:
            # if we have a custom name but no suite name, the custom name is treated
            # as our suite name *if* no image exists with the custom name
            # (this is for backwards compatibility)
            self._name = self._custom_name
            if not self.exists():
                self._suite = self._custom_name
                self._custom_name = None
        if self._custom_name == self._suite:
            self._custom_name = None

        if not self._arch:
            out, _, ret = safe_run(['dpkg', '--print-architecture'])
            if ret != 0:
                raise RuntimeError('Running dpkg --print-architecture failed: {}'.format(out))

            self._arch = out.strip()
        if self._custom_name:
            return self._custom_name
        elif self._variant:
            return '{}-{}-{}'.format(self._suite, self._variant, self._arch)
        else:
            return '{}-{}'.format(self._suite, self._arch)

    def _custom_name_parameter_check(self):
        '''Read parameters in case a custom name was passed, and perform basic sanity checks.'''
        import sys

        if self._parameters_checked:
            return
        if not self._custom_name:
            return

        config_fname = self.get_config_location()
        if not os.path.isfile(config_fname):
            print_error('No configuration data found for image "{}"!'.format(self.name))
            sys.exit(3)

        with open(config_fname, 'rt') as f:
            cdata = json.loads(f.read())

            c_suite = cdata.get('Suite', self.suite)
            if not self._suite:
                # if no suite was set, but we have one in the manifest file,
                # we will always use it to fill in the gap
                self._suite = c_suite

            if self.suite != c_suite:
                print_error(
                    'Expected suite name "{}" for image "{}", but got "{}" instead.'.format(
                        cdata.get('Suite'), self.name, self.suite
                    )
                )
                sys.exit(1)
            c_arch = cdata.get('Architecture', self.arch)
            c_variant = cdata.get('Variant', self.variant)

            if self.arch and self.arch != c_arch:
                print_warn(
                    (
                        'Expected architecture "{}" for image "{}", but got "{}" instead. '
                        'Using expected value.'
                    ).format(c_arch, self.name, self.arch)
                )
            if self.variant and self.variant != c_variant:
                print_warn(
                    (
                        'Expected variant "{}" for image "{}", but got "{}" instead. ' 'Using expected value.'
                    ).format(c_variant, self.name, self.variant)
                )
            self._arch = c_arch
            self._variant = c_variant
        self._parameters_checked = True

    @property
    def name(self) -> str:
        return self._name

    @property
    def suite(self) -> str:
        return self._suite

    @property
    def base_suite(self) -> str:
        return self._base_suite

    @property
    def arch(self) -> str:
        return self._arch

    @property
    def variant(self) -> str:
        return self._variant

    @property
    def global_config(self):
        return self._gconf

    @property
    def aptcache(self):
        return self._aptcache

    @property
    def cache_packages(self) -> bool:
        return self._gconf.cache_packages

    @property
    def has_base_suite(self) -> bool:
        return True if self.base_suite and self.base_suite != self.suite else False

    @property
    def results_dir(self):
        Path(self._results_dir).mkdir(parents=True, exist_ok=True)
        return self._results_dir

    @results_dir.setter
    def results_dir(self, path):
        self._results_dir = path
        Path(self._results_dir).mkdir(exist_ok=True)

    def _copy_helper_script(self, osroot_path):
        script_location = os.path.join(osroot_path, 'usr', 'lib', 'debspawn')
        Path(script_location).mkdir(parents=True, exist_ok=True)
        script_fname = os.path.join(script_location, 'dsrun')

        if os.path.isfile(script_fname):
            os.remove(script_fname)
        shutil.copy2(self._gconf.dsrun_path, script_fname)

        os.chmod(script_fname, 0o0755)

    def get_image_location(self):
        return os.path.join(self._gconf.osroots_dir, '{}.tar.zst'.format(self.name))

    def get_image_cache_dir(self):
        cache_img_dir = os.path.join(self._gconf.osroots_dir, 'dcache', self.name)
        Path(cache_img_dir).mkdir(parents=True, exist_ok=True)
        return cache_img_dir

    def get_cache_image_location(self):
        if not self._cachekey:
            return None
        return os.path.join(self.get_image_cache_dir(), '{}.tar.zst'.format(self._cachekey))

    def get_config_location(self):
        return os.path.join(self._gconf.osroots_dir, '{}.json'.format(self.name))

    def exists(self):
        return os.path.isfile(self.get_image_location())

    def cacheimg_exists(self):
        location = self.get_cache_image_location()
        if not location:
            return False
        return os.path.isfile(location)

    def ensure_exists(self):
        '''
        Ensure the container image exists, and terminate the
        program with an error code in case it does not.
        '''
        import sys

        if not self._load_existent():
            print_error(
                'The container image for "{}" does not exist. Please create it first.'.format(self.name)
            )
            sys.exit(3)

    def _load_existent(self) -> bool:
        '''Check if image exists, and if so load some essential data and return True.'''
        # ensure the set config values are sane if the user is using a custom container name
        if self._custom_name:
            self._custom_name_parameter_check()
        return self.exists()

    def new_nspawn_machine_name(self):
        import platform
        from random import choice
        from string import digits, ascii_lowercase

        nid = ''.join(choice(ascii_lowercase + digits) for _ in range(4))

        # on Linux, the maximum hostname length is 64, so we simple set this as general default for
        # debspawn here.
        # shorten the hostname part or replace the suffix, depending on what is longer.
        # This should only ever matter if the hostname of the system already is incredibly long
        uniq_suffix = '{}-{}'.format(self.name, nid)
        if len(uniq_suffix) > 48:
            uniq_suffix = ''.join(choice(ascii_lowercase + digits) for _ in range(12))
        node_name_prefix = platform.node()[: 63 - len(uniq_suffix)]

        return '{}-{}'.format(node_name_prefix, uniq_suffix)

    def _write_config_json(
        self, mirror, components, extra_suites, extra_source_lines, *, allow_recommends: bool, with_init: bool
    ):
        '''
        Create configuration file for this container base image
        '''

        print_info('Saving configuration settings.')
        data: T.Dict[str, T.Union[str, bool]] = {
            'Name': self.name,
            'CustomName': self._custom_name,
            'Suite': self.suite,
            'BaseSuite': self.base_suite,
            'Architecture': self.arch,
        }

        if self.variant:
            data['Variant'] = self.variant
        if mirror:
            data['Mirror'] = mirror
        if components:
            data['Components'] = components
        if extra_suites:
            data['ExtraSuites'] = extra_suites
        if extra_source_lines:
            data['ExtraSourceLines'] = extra_source_lines
        if allow_recommends:
            data['AllowRecommends'] = True
        if with_init:
            data['IncludesInit'] = True
        data['BootstrapTool'] = self._gconf.bootstrap_tool

        with open(self.get_config_location(), 'wt') as f:
            f.write(json.dumps(data, sort_keys=True, indent=4))
            f.write('\n')

    def _clear_image_tree(self, image_dir):
        '''Clear files from a directory tree that we don't want in the tarball.'''

        if os.path.ismount(image_dir):
            print_warn('Preparing OS tree for compression, but /dev is still mounted.')
            return

        for sdir, _, files in os.walk(os.path.join(image_dir, 'dev')):
            for f in files:
                fname = os.path.join(sdir, f)
                if os.path.lexists(fname) and not os.path.isdir(fname) and not os.path.ismount(fname):
                    os.remove(fname)

    def _remove_unwanted_files(self, instance_dir):
        '''Delete unwanted files from a base image'''

        # drop resolv.conf: Some OSes set this to a symlink, which will lead to nowhere
        # in the container and may cause issues when bindmounting over it
        maybe_remove(os.path.join(instance_dir, 'etc', 'resolv.conf'))

        # our APT proxy connfiguration should also not be stored, we will reset it every time
        maybe_remove(os.path.join(instance_dir, 'etc', 'apt', 'apt.conf.d', '98debspawn_proxy'))

        # drop machine-id file, if one exists
        maybe_remove(os.path.join(instance_dir, 'etc', 'machine-id'))
        # drop logfiles which we want to reset
        maybe_remove(os.path.join(instance_dir, 'var', 'log', 'lastlog'))
        maybe_remove(os.path.join(instance_dir, 'var', 'log', 'faillog'))

        # drop APT package caches, we do not want them in the tarballs
        apt_pkg_cache_dir = os.path.join(instance_dir, 'var', 'cache', 'apt', 'archives')
        if os.path.isdir(apt_pkg_cache_dir):
            rmtree_mntsafe(apt_pkg_cache_dir, ignore_errors=True)

    def _setup_apt_proxy(self, instance_dir):
        '''Setup APT proxy configuration.
        APT needs special treatment even in the container to work with a company proxy (yuck!),
        so we do this here so the user does not need to care about it.
        '''

        http_proxy = os.getenv('HTTP_PROXY', os.getenv('http_proxy'))
        https_proxy = os.getenv('HTTPS_PROXY', os.getenv('https_proxy'))
        if not http_proxy and not https_proxy:
            return
        if http_proxy:
            http_proxy = http_proxy.replace('"', '')
        if https_proxy:
            https_proxy = https_proxy.replace('"', '')

        proxyconf_fname = os.path.join(instance_dir, 'etc', 'apt', 'apt.conf.d', '98debspawn_proxy')
        with open(proxyconf_fname, 'w') as f:
            if http_proxy:
                f.write('Acquire::http::Proxy "{}";\n'.format(http_proxy))
            if https_proxy:
                f.write('Acquire::https::Proxy "{}";\n'.format(https_proxy))

    def _setup_apt_repo_preferences(self, instance_dir, preferred_suites):
        '''Setup APT repository preferences.
        APT somtimes wants to install packages from older repositories to satisfy
        dependencies, even though we must build with versions from the newer repositories.
        This especially is the case for `NotAutomatic` suites like "experimental", where packages
        would not get auto-installed from.
        Passing "-t <suite>" to the APT commands restricts the solver way too much, so
        valid dependency chains are no chosen for partial suites if the suites don't
        contain all dependencies as well.
        This is a compromise to make APT do the right thing, by simply setting all suites
        to the same priority.
        NOTE: With suites like experimental, we may need to set the preferences later if we want
        to avoid experimental packages being added to the base image. That comes with its own
        difficulties though, so we avoid that for now.
        '''

        # enusre all suites are represented in the "preferred suites" list
        if not preferred_suites:
            preferred_suites = []
        if self.has_base_suite:
            if self.base_suite not in preferred_suites:
                preferred_suites.insert(0, self.base_suite)
        if self.suite not in preferred_suites:
            preferred_suites.insert(0, self.suite)

        aptpref_fname = os.path.join(instance_dir, 'etc', 'apt', 'preferences.d', '10debspawn')
        with open(aptpref_fname, 'w') as f:
            first = True
            for suite in preferred_suites:
                if not first:
                    f.write('\n')
                first = False
                priority = 500
                if suite == self.suite:
                    priority = 600
                f.write(('Package: *\n' 'Pin: release a={}\n' 'Pin-Priority: {}\n').format(suite, priority))

    def _create_internal(
        self,
        mirror=None,
        components=None,
        *,
        extra_suites: list[str] = None,
        extra_source_lines: str = None,
        allow_recommends: bool = False,
        with_init: bool = False,
        show_header: bool = True,
    ):
        '''Create new container base image (internal method)'''

        if self.exists():
            print_error('An image already exists for this configuration. Can not create a new one.')
            return False
        if not extra_suites:
            extra_suites = []

        bootstrap_tool_exe = self._gconf.bootstrap_tool
        if not shutil.which(bootstrap_tool_exe):
            print_error('Unable to find executable for bootstrap tool "{}".'.format(bootstrap_tool_exe))
            return False

        # ensure image location exists
        Path(self._gconf.osroots_dir).mkdir(parents=True, exist_ok=True)

        if show_header:
            print_header('Creating new base: {} [{}]'.format(self.suite, self.arch))
        else:
            print_section('Creating new base: {} [{}]'.format(self.suite, self.arch))

        print('Bootstrap tool:', '{} {}'.format(bootstrap_tool_exe, bootstrap_tool_version(self._gconf)))
        if bootstrap_tool_exe not in self._known_bootstrap_tools:
            print_warn('Using unfamiliar bootstrap tool: {}'.format(bootstrap_tool_exe))
        if self._custom_name:
            print('Custom name: {}'.format(self.name))
        print('Using mirror: {}'.format(mirror if mirror else 'default'))
        if self.variant:
            print('Variant: {}'.format(self.variant))
        if with_init:
            print('Includes init: yes')

        include_pkgs = ['python3-minimal', 'eatmydata']
        if with_init:
            include_pkgs.append('systemd-sysv')
        cmd = [bootstrap_tool_exe, '--arch={}'.format(self.arch), '--include=' + ','.join(include_pkgs)]
        if components:
            cmd.append('--components={}'.format(','.join(components)))
        if self.variant:
            cmd.append('--variant={}'.format(self.variant))

        with temp_dir() as tdir:
            bootstrap_suite = self.suite
            if self.has_base_suite:
                bootstrap_suite = self.base_suite
            cmd.extend([bootstrap_suite, tdir])
            print('Bootstrap suite: {}'.format(bootstrap_suite))
            if extra_suites:
                print('Additional suites: {}'.format(', '.join(extra_suites)))
            if extra_source_lines:
                print('Custom sources.list lines will be added:')
                for line in extra_source_lines.split('\\n'):
                    print('    {}'.format(line))
            if mirror:
                cmd.append(mirror)

            print_section('Bootstrap')
            proc = subprocess.run(cmd, check=False)
            if proc.returncode != 0:
                return False

            # create helper script runner
            self._copy_helper_script(tdir)

            # if we bootstrapped the base suite, add the primary suite to
            # sources.list. We also add any explicit extra suites and source lines
            if self.has_base_suite or extra_suites or extra_source_lines:
                import re

                sourceslist_fname = os.path.join(tdir, 'etc', 'apt', 'sources.list')
                if not mirror:
                    with open(sourceslist_fname, 'r') as f:
                        contents = f.read()
                        matches = re.findall(
                            'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+',
                            contents,
                        )
                        if not matches:
                            print_error('Unable to detect default APT repository URL (no regex matches).')
                            return False
                        mirror = matches[0]
                    if not mirror:
                        print_error('Unable to detect default APT repository URL.')
                        return False

                if not components:
                    # FIXME: We should really be more clever here, e.g. depend on python-apt
                    # and parse sources.list properly
                    components = ['main']
                with open(sourceslist_fname, 'a') as f:
                    if self.has_base_suite:
                        f.write(
                            'deb {mirror} {suite} {components}\n'.format(
                                mirror=mirror, suite=self.suite, components=' '.join(components)
                            )
                        )

                    if extra_suites:
                        f.write('\n')
                        for esuite in extra_suites:
                            if esuite == self.suite or esuite == bootstrap_suite:
                                # don't add existing suites multiple times
                                continue
                            f.write(
                                'deb {mirror} {esuite} {components}\n'.format(
                                    mirror=mirror, esuite=esuite, components=' '.join(components)
                                )
                            )

                    if extra_source_lines:
                        f.write('\n')
                        for line in extra_source_lines.split('\\n'):
                            f.write('{}\n'.format(line.strip()))

            # set preference suites in dependency resolution
            self._setup_apt_repo_preferences(tdir, extra_suites)

            # write our default APT settings for this container
            aptconf_fname = os.path.join(tdir, 'etc', 'apt', 'apt.conf.d', '99debspawn')
            with open(aptconf_fname, 'w') as f:
                # fail immediately with a proper exit code when e.g. apt update fails,
                # so we can retry and don't silently use old packages
                # (only available with newer APT versions)
                f.write('APT::Update::Error-Mode "any";\n')
                f.write('APT::AutoRemove::SuggestsImportant "false";\n')
                f.write('APT::AutoRemove::RecommendsImportant "false";\n')
                f.write('Acquire::Languages "none";\n')
                if not allow_recommends:
                    f.write('APT::Install-Recommends "false";\n')
                    f.write('APT::Install-Suggests "false";\n')

            if with_init:
                # if we are allowing to boot this container, add a passwordless root login
                # hack so the user gets a root shell instead of a login prompt
                getty_override_fname = os.path.join(
                    tdir, 'etc', 'systemd', 'system', 'console-getty.service.d', 'autorootlogin.conf'
                )
                os.makedirs(os.path.dirname(getty_override_fname), exist_ok=True)
                with open(getty_override_fname, 'w') as f:
                    # fail immediately with a proper exit code when e.g. apt update fails,
                    # so we can retry and don't silently use old packages
                    # (only available with newer APT versions)
                    f.write(
                        (
                            '[Service]\n'
                            'ExecStart=\n'
                            'ExecStartPre=-/usr/bin/sed -i \'/pam_loginuid.so/d\' /etc/pam.d/login\n'
                            'ExecStart=-/sbin/agetty --autologin root --noclear '
                            '--keep-baud - 115200,38400,9600 $TERM\n'
                        )
                    )

            # delete unwanted files, especially resolv.conf as a broken one will
            # mess with the next step
            self._remove_unwanted_files(tdir)

            # configure APT proxy, so the configure operation will work behind proxys
            self._setup_apt_proxy(tdir)

            print_section('Configure')
            if (
                nspawn_run_helper_persist(
                    self, tdir, self.new_nspawn_machine_name(), '--update', build_uid=self._builder_uid
                )
                != 0
            ):
                return False

            # drop any unwanted files (again) before building the tarball
            self._remove_unwanted_files(tdir)

            print_section('Creating Tarball')
            self._clear_image_tree(tdir)
            compress_directory(tdir, self.get_image_location())

        # store configuration settings, so we can later recreate this tarball
        # or just display information about it
        self._write_config_json(
            mirror,
            components,
            extra_suites,
            extra_source_lines,
            allow_recommends=allow_recommends,
            with_init=with_init,
        )

        return True

    def create(
        self,
        mirror: str = None,
        components: list[str] = None,
        *,
        extra_suites: list[str] = None,
        extra_source_lines: str = None,
        allow_recommends: bool = False,
        with_init: bool = False,
    ):
        '''Create new container base image (internal method)'''
        ensure_root()

        if self.exists():
            print_error('This configuration has already been created. You can only delete or update it.')
            return False

        ret = self._create_internal(
            mirror=mirror,
            components=components,
            extra_suites=extra_suites,
            extra_source_lines=extra_source_lines,
            allow_recommends=allow_recommends,
            with_init=with_init,
            show_header=True,
        )
        if ret:
            print_info('Done.')

        return ret

    def delete(self):
        '''Remove container base image'''
        ensure_root()

        if not self._load_existent():
            print_error('Can not delete "{}": This configuration does not exist.'.format(self.name))
            return False

        print_header('Removing base image {}'.format(self.name))

        print_section('Deleting cache')
        # remove packages cache
        cache_size = self._aptcache.clear()
        print_info('Removed {} cached packages.'.format(cache_size))
        self._aptcache.delete()
        # remove cached images
        shutil.rmtree(self.get_image_cache_dir())
        print_info('Cache directory removed.')

        print_section('Deleting base tarball')
        os.remove(self.get_image_location())

        config_fname = self.get_config_location()
        if os.path.isfile(config_fname):
            print_section('Deleting configuration manifest')
            os.remove(config_fname)

        print_info('Done.')
        return True

    @contextmanager
    def new_instance(self, basename=None):
        with temp_dir() as tdir:
            if self.cacheimg_exists():
                image_fname = self.get_cache_image_location()
            else:
                image_fname = self.get_image_location()

            decompress_tarball(image_fname, tdir)
            self._setup_apt_proxy(tdir)

            yield tdir, self.new_nspawn_machine_name()

    def make_instance_permanent(self, instance_dir):
        '''Add changes done in the current instance to the main tarball of this OS tree, replacing it.'''

        # remove unwanted files from the tarball
        self._clear_image_tree(instance_dir)

        if self._cachekey:
            tarball_name = self.get_cache_image_location()
        else:
            tarball_name = self.get_image_location()
        tarball_name_old = '{}.old'.format(tarball_name)

        if os.path.isfile(tarball_name):
            os.replace(tarball_name, tarball_name_old)
        compress_directory(instance_dir, tarball_name)
        if os.path.isfile(tarball_name_old):
            os.remove(tarball_name_old)

        tar_size = os.path.getsize(tarball_name)
        if self._cachekey:
            print_info(
                'New compressed tarball size (for {}) is {}'.format(self._cachekey, format_filesize(tar_size))
            )
        else:
            print_info('New compressed tarball size is {}'.format(format_filesize(tar_size)))

    def update(self):
        '''Update container base image'''
        ensure_root()

        if not self._load_existent():
            print_error('Can not update "{}": This configuration does not exist.'.format(self.name))
            return False

        print_header('Updating container image')

        with self.new_instance() as (instance_dir, _):
            # ensure helper script runner exists and is up to date
            self._copy_helper_script(instance_dir)

            print_section('Update')
            if (
                nspawn_run_helper_persist(
                    self,
                    instance_dir,
                    self.new_nspawn_machine_name(),
                    '--update',
                    build_uid=self._builder_uid,
                )
                != 0
            ):
                return False

            # drop unwanted files from the image
            self._remove_unwanted_files(instance_dir)

            print_section('Recreating tarball')
            self.make_instance_permanent(instance_dir)

        print_section('Cleaning up cache')
        cache_size = self._aptcache.clear()
        print_info('Removed {} cached packages.'.format(cache_size))
        # remove now-outdated cached images
        shutil.rmtree(self.get_image_cache_dir())

        print_info('Done.')
        return True

    def recreate(self):
        '''Recreate a container base image'''
        ensure_root()

        if not self._load_existent():
            print_error('Can not recreate "{}": The image does not exist.'.format(self.name))
            return False

        config_fname = self.get_config_location()
        if not os.path.isfile(config_fname):
            print_error(
                'Can not recreate "{}": Unable to find configuration data for this image.'.format(self.name)
            )
            return False

        print_header('Recreating container image')

        # read configuration data
        with open(config_fname, 'rt') as f:
            cdata: T.Dict[str, T.Union[str, bool]] = json.loads(f.read())
            self._name = cdata.get('Name', self.name)
            self._custom_name = cdata.get('CustomName', self._custom_name)
            self._suite = cdata.get('Suite', self.suite)
            self._base_suite = cdata.get('BaseSuite', self.base_suite)
            self._arch = cdata.get('Architecture', self.arch)
            self._variant = cdata.get('Variant', self.variant)
            mirror = cdata.get('Mirror')
            components = cdata.get('Components')
            extra_suites = cdata.get('ExtraSuites', [])
            extra_source_lines = cdata.get('ExtraSourceLines')
            allow_recommends = cdata.get('AllowRecommends', False)
            with_init = cdata.get('IncludesInit', False)

        print_section('Deleting cache')
        cache_size = self._aptcache.clear()
        print_info('Removed {} cached packages.'.format(cache_size))
        self._aptcache.delete()
        print_info('Cache directory removed.')

        # move old image tarball out of the way
        image_name = self.get_image_location()
        image_name_old = self.get_image_location() + '.old'
        if os.path.isfile(image_name_old):
            print_info('Removing cruft image')
            os.remove(image_name_old)
        os.rename(image_name, image_name_old)
        print_info('Old tarball moved.')

        # ty to create the tarball again
        try:
            ret = self._create_internal(
                mirror=mirror,
                components=components,
                extra_suites=extra_suites,
                extra_source_lines=extra_source_lines,
                allow_recommends=allow_recommends,
                with_init=with_init,
                show_header=False,
            )
        except Exception as e:
            print_error('Error while trying to create image: {}'.format(str(e)))
            ret = False

        if ret:
            if os.path.isfile(image_name_old):
                print_info('Removing old image')
                os.remove(image_name_old)

            print_info('Removing outdated cached images')
            shutil.rmtree(self.get_image_cache_dir())

            print_info('Done.')
            return True
        else:
            print_info('Restoring old tarball')
            if os.path.isfile(image_name):
                print_info('Removing failed new image')
                os.remove(image_name)
            os.rename(image_name_old, image_name)
            print_info('Recreation failed.')
            return False

    def login(self, persistent=False, *, allowed: list[str] = None, boot: bool = False):
        '''Interactive shell login into the container'''
        ensure_root()

        if not self._load_existent():
            print_info('Can not enter "{}": This configuration does not exist.'.format(self.name))
            return False

        print_header(
            'Login (persistent changes) for {}'.format(self.name)
            if persistent
            else 'Login for {}'.format(self.name)
        )
        with self.new_instance() as (instance_dir, _):
            # ensure helper script runner exists and is up to date
            self._copy_helper_script(instance_dir)

            # run an interactive shell in the new container
            nspawn_run_persist(
                self,
                instance_dir,
                self.new_nspawn_machine_name(),
                '/srv',
                verbose=True,
                allowed=allowed,
                boot=boot,
            )

            if persistent:
                print_section('Recreating tarball')
                self.make_instance_permanent(instance_dir)
            else:
                print_info('Changes discarded.')

        print_info('Done.')
        return True

    def retrieve_artifacts(self, src_dir: str, dest_dir: T.Optional[str] = None):
        from glob import glob

        print_section('Retrieving build artifacts')
        if not dest_dir:
            dest_dir = self.results_dir

        o_uid, o_gid = get_owner_uid_gid()
        acount = 0
        for f in glob(os.path.join(src_dir, '*.*')):
            if os.path.isfile(f):
                target_fname = os.path.join(dest_dir, os.path.basename(f))
                safe_copy(f, target_fname)
                os.chown(target_fname, o_uid, o_gid, follow_symlinks=False)
                acount += 1
        print_info('Copied {} files.'.format(acount))

    def _copy_command_script_to_instance_dir(self, instance_dir: str, command_script: str) -> T.Optional[str]:
        '''
        Copy a script from the host to the current instance directory and make it
        executable.
        Contains the path to the executable script as seen from inside the container.
        '''
        host_script = os.path.abspath(command_script)
        if not os.path.isfile(host_script):
            return None

        script_location = os.path.join(instance_dir, 'srv', 'tmp')
        Path(script_location).mkdir(parents=True, exist_ok=True)
        script_fname = os.path.join(script_location, os.path.basename(host_script))

        if os.path.isfile(script_fname):
            os.remove(script_fname)
        shutil.copy2(host_script, script_fname)
        os.chmod(script_fname, 0o0755)

        return os.path.join('/srv', 'tmp', os.path.basename(host_script))

    def run(
        self,
        command,
        build_dir,
        artifacts_dir,
        boot: bool = False,
        init_command=None,
        copy_command=False,
        header_msg=None,
        bind_build_dir: T.Optional[str] = None,
        allowed: list[str] = None,
    ):
        '''Run an arbitrary command or script in the container'''
        ensure_root()

        if not self._load_existent():
            print_error('Can not run command in "{}": The base image does not exist.'.format(self.name))
            return False

        if len(command) <= 0:
            print_error('No command was given. Can not continue.')
            return False

        if isinstance(init_command, str):
            if init_command:
                import shlex

                init_command = shlex.split(init_command)
        init_command = listify(init_command)
        allowed = listify(allowed)
        if bind_build_dir == 'n':
            bind_build_dir = None

        # ensure we have absolute paths
        if build_dir:
            build_dir = os.path.normpath(os.path.abspath(build_dir))
        if artifacts_dir:
            artifacts_dir = os.path.normpath(os.path.abspath(artifacts_dir))

        if self._cachekey and init_command and not self.cacheimg_exists():
            print_header('Preparing template for `{}`'.format(self._cachekey))

            # we do not have a cached image prepared, let's do that now!
            with self.new_instance() as (instance_dir, machine_name):
                # ensure helper script runner exists and is up to date
                self._copy_helper_script(instance_dir)

                if copy_command:
                    # copy initialization script from host to container
                    host_script = init_command[0]
                    init_command[0] = self._copy_command_script_to_instance_dir(instance_dir, host_script)
                    if not init_command[0]:
                        print_error(
                            (
                                'Unable to find initialization script "{}", '
                                'can not copy it to the container. Exiting.'
                            ).format(host_script)
                        )
                        return False

                r = nspawn_run_helper_persist(
                    self, instance_dir, machine_name, '--prepare-run', '/srv', build_uid=self._builder_uid
                )
                if r != 0:
                    print_error('Container setup failed.')
                    return False
                # we do not want some permissions to be in effect here,
                # as they may have unwanted effects on the final cached image
                banned_permissions = ['full-dev', 'full-proc', 'read-kmods']
                filtered_allowed = []
                for perm in allowed:
                    if perm not in banned_permissions:
                        filtered_allowed.append(perm)

                init_nspawn_flags = []
                if build_dir:
                    if not bind_build_dir:
                        shutil.copytree(
                            build_dir, os.path.join(instance_dir, 'srv', 'build'), dirs_exist_ok=True
                        )
                    else:
                        if bind_build_dir == 'rw':
                            init_nspawn_flags = ['--bind={}:/srv/build/'.format(build_dir)]
                        elif bind_build_dir == 'ro':
                            init_nspawn_flags = ['--bind-ro={}:/srv/build/'.format(build_dir)]
                r = nspawn_run_persist(
                    self,
                    instance_dir,
                    machine_name,
                    '/srv',
                    init_command,
                    init_nspawn_flags,
                    allowed=filtered_allowed,
                )
                if r != 0:
                    return False

                print_info('Storing prepared image in cache')
                self.make_instance_permanent(instance_dir)

        if header_msg:
            print_header(header_msg)
        if self._cachekey and init_command and self.cacheimg_exists():
            print_info('Using cached container image `{}`'.format(self._cachekey))

        with self.new_instance() as (instance_dir, machine_name):
            # ensure helper script runner exists and is up to date
            self._copy_helper_script(instance_dir)

            if copy_command:
                # copy the script from the host into our container and execute it there
                host_script = command[0]
                command[0] = self._copy_command_script_to_instance_dir(instance_dir, host_script)
                if not command[0]:
                    print_error(
                        ('Unable to find script "{}", can not copy it to the container. ' 'Exiting.').format(
                            host_script
                        )
                    )
                    return False

            r = nspawn_run_helper_persist(
                self, instance_dir, machine_name, '--prepare-run', '/srv', build_uid=self._builder_uid
            )
            if r != 0:
                print_error('Container setup failed.')
                return False

            # Create a few directories we may use for bindmounting if some allow-flags are set,
            # and which are not commonly present in the base image.
            # This is only needed for `run` actions, and regular package builds should not require
            # bindmounts to these directories.
            os.makedirs(os.path.join(instance_dir, 'lib', 'modules'), exist_ok=True)
            os.makedirs(os.path.join(instance_dir, 'boot'), exist_ok=True)
            os.makedirs(os.path.join(instance_dir, 'srv', 'artifacts'), exist_ok=True)

            print_section('Running Task')
            nspawn_flags = []
            chdir = '/srv'
            if build_dir:
                chdir = '/srv/build'
                if not bind_build_dir:
                    shutil.copytree(build_dir, os.path.join(instance_dir, 'srv', 'build'), dirs_exist_ok=True)
                else:
                    if bind_build_dir == 'rw':
                        nspawn_flags.extend(['--bind={}:/srv/build/'.format(build_dir)])
                    elif bind_build_dir == 'ro':
                        nspawn_flags.extend(['--bind-ro={}:/srv/build/'.format(build_dir)])

            r = nspawn_run_persist(
                self, instance_dir, machine_name, chdir, command, nspawn_flags, allowed=allowed, boot=boot
            )
            if r != 0:
                return False

            # copy results to target directory
            self.retrieve_artifacts(os.path.join(instance_dir, 'srv', 'artifacts'), artifacts_dir)

        print_info('Done.')
        return True


def print_container_base_image_info(gconf):
    '''
    Search for all available container base images and list information
    about them.
    '''
    from glob import glob

    osroots_dir = gconf.osroots_dir
    tar_files = []
    if os.path.isdir(osroots_dir):
        tar_files = list(glob(os.path.join(osroots_dir, '*.tar.zst')))
    if not tar_files:
        print_info('No container base images have been found!')
        return False
    tar_files_len = len(tar_files)

    for i, tar_fname in enumerate(tar_files):
        img_basepath = os.path.splitext(os.path.splitext(tar_fname)[0])[0]
        config_fname = img_basepath + '.json'
        imgid = os.path.basename(img_basepath)
        print('[{}]'.format(imgid))

        cache_files = list(glob(os.path.join(osroots_dir, 'dcache', imgid, '*.tar.zst')))
        cached_names = []
        for cfile in cache_files:
            cname = os.path.basename(os.path.splitext(os.path.splitext(cfile)[0])[0])
            cached_names.append(cname)

        # read configuration data if it exists
        if os.path.isfile(config_fname):
            with open(config_fname, 'rt') as f:
                cdata = json.loads(f.read())
            for key, value in cdata.items():
                if type(value) is list:
                    value = '; '.join(value)
                print('{} = {}'.format(key, value))

        tar_size = os.path.getsize(tar_fname)
        print('Size = {}'.format(format_filesize(tar_size)))

        if cached_names:
            print('CachedImages = {}'.format('; '.join(cached_names)))
        if i != tar_files_len - 1:
            print()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2022 Matthias Klumpp <matthias@tenstral.net>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import os
import sys
from functools import reduce
from subprocess import check_call
from xml.sax.saxutils import escape as xml_escape

sys.path.append("..")


class DocbookEditor:
    def __init__(self):
        self._replacements = {}

    def add_substvar(self, name, replacement):
        self._replacements['@{}@'.format(name)] = replacement

    def register_command_flag_synopsis(self, actions, command_name):
        flags_text = ''
        flags_entries = ''
        for item in actions:
            options_text = xml_escape('|'.join(item.option_strings))
            flags_text += '<arg>{}</arg>\n'.format(options_text)

            oid = item.option_strings[0]
            desc_text = None
            if oid == '-h':
                desc_text = 'Print brief help information about available commands.'
            if command_name != 'create':
                if oid == '--variant':
                    desc_text = 'Set the variant of the selected image, that was used for bootstrapping.'
                elif oid == '-a':
                    desc_text = 'The architecture of the base image that should be selected.'

            if not desc_text:
                desc_text = item.help
            desc_text = xml_escape(desc_text)

            if desc_text.startswith('CF|'):
                desc_text = desc_text[3:]
                desc_text = desc_text.replace('binary:', '<option>binary</option>:', 1)
                desc_text = desc_text.replace('arch:', '<option>arch</option>:', 1)
                desc_text = desc_text.replace('indep:', '<option>indep</option>:', 1)
                desc_text = desc_text.replace('source:', '<option>source</option>:', 1)

            flags_entries += '''<varlistentry>
                                  <term>{}</term>
                                  <listitem>
                                      <para>
                                      {}
                                      </para>
                                 </listitem>
                              </varlistentry>'''.format(
                options_text, desc_text
            )

        self.add_substvar('{}_FLAGS_SYNOPSIS'.format(command_name.upper()), flags_text)
        self.add_substvar('{}_FLAGS_ENTRIES'.format(command_name.upper()), flags_entries)

    def process_file(self, input_fname, output_fname):
        with open(input_fname, 'r') as f:
            template_content = f.read()

        result = reduce(
            lambda x, y: x.replace(y, self._replacements[y]), self._replacements, template_content
        )

        with open(output_fname, 'w') as f:
            f.write(result)

        return output_fname


def generate_docbook_pages(build_dir):
    from debspawn.cli import create_parser

    build_dir = os.path.abspath(build_dir)

    parser = create_parser()
    editor = DocbookEditor()
    editor.register_command_flag_synopsis(parser._get_optional_actions(), 'BASE')

    xml_manpages = []
    xml_manpages.append(editor.process_file('docs/debspawn.1.xml', os.path.join(build_dir, 'debspawn.1.xml')))

    for command, sp in parser._get_positional_actions()[0]._name_parser_map.items():
        editor.register_command_flag_synopsis(sp._get_optional_actions(), command)
        template_fname = 'docs/debspawn-{}.1.xml'.format(command)
        if not os.path.isfile(template_fname):
            if command in ['ls', 'b']:
                continue  # the ls and b shorthands need to manual page
            print('Manual page template {} is missing! Skipping it.'.format(template_fname))
            continue

        xml_manpages.append(
            editor.process_file(template_fname, os.path.join(build_dir, os.path.basename(template_fname)))
        )

    return xml_manpages


def create_manpage(xml_src, out_dir):
    man_name = os.path.splitext(os.path.basename(xml_src))[0]
    out_fname = os.path.join(out_dir, man_name)

    print('Generating manual page {}'.format(man_name))
    check_call(
        [
            'xsltproc',
            '--nonet',
            '--stringparam',
            'man.output.quietly',
            '1',
            '--stringparam',
            'funcsynopsis.style',
            'ansi',
            '--stringparam',
            'man.th.extra1.suppress',
            '1',
            '-o',
            out_fname,
            'http://docbook.sourceforge.net/release/xsl/current/manpages/docbook.xsl',
            xml_src,
        ]
    )
    return out_fname


if __name__ == '__main__':
    generate_docbook_pages('/tmp')



#!/usr/bin/env python3

import os
import sys
import shutil
import platform
from subprocess import check_call

from setuptools import setup
from setuptools.command.install_scripts import install_scripts as install_scripts_orig

sys.path.append(os.getcwd())
from debspawn import __appname__, __version__  # noqa: E402

thisfile = __file__
if not os.path.isabs(thisfile):
    thisfile = os.path.normpath(os.path.join(os.getcwd(), thisfile))
source_root = os.path.dirname(thisfile)


class install_scripts(install_scripts_orig):
    def _check_command(self, command):
        if not shutil.which(command):
            print(
                'The "{}" binary was not found. Please install it to continue!'.format(command),
                file=sys.stderr,
            )
            sys.exit(1)

    def _check_commands_available(self):
        '''Check if certain commands are available that debspawn needs to work.'''
        self._check_command('systemd-nspawn')
        self._check_command('findmnt')
        self._check_command('zstd')
        self._check_command('debootstrap')
        self._check_command('dpkg')

    def run(self):
        if platform.system() == 'Windows':
            super().run()
            return

        if not self.skip_build:
            self.run_command('build_scripts')
        self.outfiles = []

        if self.dry_run:
            return

        # We want the files to be installed without a suffix on Unix
        self.mkpath(self.install_dir)
        for infile in self.get_inputs():
            infile = os.path.basename(infile)
            in_built = os.path.join(self.build_dir, infile)
            in_stripped = infile[:-3] if infile.endswith('.py') else infile
            outfile = os.path.join(self.install_dir, in_stripped)
            # NOTE: Mode is preserved by default
            self.copy_file(in_built, outfile)
            self.outfiles.append(outfile)

        # try to install configuration snippets, manual pages and other external data
        bin_install_dir = str(self.install_dir)
        if '/usr/' in bin_install_dir:
            install_root = bin_install_dir.split('/usr/', 1)[0]
            prefix = '/usr/local' if '/usr/local/' in bin_install_dir else '/usr'
            sysdata_install_script = os.path.join(source_root, 'install-sysdata.py')
            if os.path.isfile(sysdata_install_script) and os.path.isdir(install_root):
                check_call(
                    [sys.executable, sysdata_install_script, '--root', install_root, '--prefix', prefix]
                )
            else:
                print('Unable to install externally managed data!', file=sys.stderr)
        else:
            print(
                (
                    '\n\n ------------------------\n'
                    'Unable to install external configuration and manual pages!\n'
                    'While these files are not essential to work with debspawn, they will improve how it runs '
                    'or are useful as documentation. Please install these files manually by running the '
                    '`install-sysdata.py` script from debspawn\'s source directory manually as root.\n'
                    'Installing these external files is not possible when installing e.g. with pip. If `setup.py` is '
                    'used directly we make an attempt to install the files, but this attempt has failed.'
                    '\n ------------------------\n\n'
                ),
                file=sys.stderr,
            )


cmdclass = {
    'install_scripts': install_scripts,
}

packages = [
    'debspawn',
    'debspawn.utils',
]

package_data = {'': ['debspawn/dsrun']}

scripts = ['debspawn.py']

install_requires = ['tomlkit>=0.8']

setup(
    name=__appname__,
    version=__version__,
    author="Matthias Klumpp",
    author_email="matthias@tenstral.net",
    description='Easily build Debian packages in systemd-nspawn containers',
    license="LGPL-3.0+",
    url="https://github.com/lkhq/debspawn",
    long_description=open(os.path.join(source_root, 'README.md')).read(),
    long_description_content_type='text/markdown',
    #
    python_requires='>=3.9',
    platforms=['any'],
    zip_safe=False,
    include_package_data=True,
    #
    packages=packages,
    cmdclass=cmdclass,
    package_data=package_data,
    scripts=scripts,
    install_requires=install_requires,
)
